""" date_utils.py: Utility functions for working with dates.

Author
    Curtis McCully (cmccully@lcogt.net)

October 2015
"""
from __future__ import absolute_import, print_function, division

import datetime
import numpy as np



def epoch_string_to_date(epoch):
    """
    Convert a string in the form of "YYYYMMDD" into a datetime object

    Parameters
    ----------
    epoch : string
        String to convert to a list of datetime objects.
        The string must have the format of YYYYMMDD.

    Returns
    -------
    datetime_epoch : datetime
    """
    return datetime.date(int(epoch[0:4]), int(epoch[4:6]), int(epoch[6:8]))


def epoch_date_to_string(epoch):
    """
    Convert a datetime object to a string with the format YYYYMMDD

    Parameters
    ----------
    epoch : datetime object or string
        datetime object to convert to a string

    Returns
    -------
    epoch_string : string
        Epoch string in the format of YYYYMMDD

    Notes
    -----
    If the input epoch is already string, then it is just returned.
    """
    return str(epoch).replace('-', '')


def parse_epoch_string(epoch_string):
    """
    Parse a string into a list of datetime objects

    Parameters
    ----------
    epoch_string : string
        String to convert to a list of datetime objects.
        The string can have the format of YYYYMMDD or YYYYMMDD-YYYYMMDD

    Returns
    -------
    epoch_list : list of datetime objects
    """
    if '-' in epoch_string:
        epoch1, epoch2 = epoch_string.split('-')
        start = epoch_string_to_date(epoch1)
        stop = epoch_string_to_date(epoch2)

        epoch_list = []
        for i in range((stop - start).days + 1):
            epoch = start + datetime.timedelta(days=i)
            epoch_list.append(str(epoch).replace('-', ''))
    else:
        epoch_list = [epoch_string]

    return epoch_list


def parse_date_obs(date_obs_string):
    # Check if there is fractional seconds in the time
    date_fractional_seconds_list = date_obs_string.split('.')
    if len(date_fractional_seconds_list) > 1:
        # Pad the string with zeros to include microseconds
        date_obs_string += (6 - len(date_fractional_seconds_list[-1])) * '0'
    else:
        # Pad the string with zeros to include microseconds
        date_obs_string += '.000000'
    return datetime.datetime.strptime(date_obs_string, '%Y-%m-%dT%H:%M:%S.%f')


def date_obs_to_string(date_obs):
    return date_obs.strftime('%Y-%m-%dT%H:%M:%S.%f')


def mean_date(dates):
    time_offsets = np.array([d - min(dates) for d in dates])
    average_offset = time_offsets.sum() / time_offsets.size
    return min(dates) + average_offset



from __future__ import absolute_import, print_function, division

from banzai import logs
from banzai.utils import fits_utils
from banzai.stages import Stage

import os

__author__ = 'cmccully'


def _trim_image(image):
    trimsec = fits_utils.parse_region_keyword(image.header['TRIMSEC'])

    image.data = image.data[trimsec]
    image.bpm = image.bpm[trimsec]

    # Update the NAXIS and CRPIX keywords
    image.header['NAXIS1'] = trimsec[1].stop - trimsec[1].start
    image.header['NAXIS2'] = trimsec[0].stop - trimsec[0].start
    image.header['CRPIX1'] -= trimsec[1].start
    image.header['CRPIX2'] -= trimsec[0].start

    image.header['L1STATTR'] = (1, 'Status flag for overscan trimming')

    return image.header['NAXIS1'], image.header['NAXIS2']


class Trimmer(Stage):
    def __init__(self, pipeline_context):
        super(Trimmer, self).__init__(pipeline_context)

    @property
    def group_by_keywords(self):
        return None

    def do_stage(self, images):

        for image in images:

            logging_tags = logs.image_config_to_tags(image, self.group_by_keywords)
            logs.add_tag(logging_tags, 'filename', os.path.basename(image.filename))
            logs.add_tag(logging_tags, 'trimsec', image.header['TRIMSEC'])
            self.logger.info('Trimming image', extra=logging_tags)

            nx, ny = _trim_image(image)
            image.update_shape(nx, ny)
        return images



from __future__ import absolute_import, print_function, division

import os
import itertools

from sqlalchemy.sql import func

from . import dbs
from .utils import date_utils
from . import logs
from banzai.images import check_image_homogeneity, Image

import abc

logger = logs.get_logger(__name__)

__author__ = 'cmccully'


class Stage(object):
    __metaclass__ = abc.ABCMeta

    @property
    def stage_name(self):
        return '.'.join([__name__, self.__class__.__name__])

    @abc.abstractproperty
    def group_by_keywords(self):
        pass

    def __init__(self, pipeline_context):
        self.logger = logs.get_logger(self.stage_name)
        self.pipeline_context = pipeline_context

    def get_grouping(self, image):
        grouping_criteria = [image.site, image.instrument, image.epoch]
        if self.group_by_keywords:
            grouping_criteria += [image.header[keyword] for keyword in self.group_by_keywords]
        return grouping_criteria

    def run_stage(self, image_set, image_config):
        image_set = list(image_set)
        tags = logs.image_config_to_tags(image_set[0], self.group_by_keywords)
        self.logger.info('Running {0}'.format(self.stage_name), extra=tags)
        return self.do_stage(image_set)

    def run(self, images):
        images.sort(key=self.get_grouping)
        processed_images = []
        for image_config, image_set in itertools.groupby(images, self.get_grouping):
            try:
                processed_images += self.run_stage(image_set, image_config)
            except MasterCalibrationDoesNotExist as e:
                continue
        return processed_images

    @abc.abstractmethod
    def do_stage(self, images):
        pass


class CalibrationMaker(Stage):
    __metaclass__ = abc.ABCMeta

    def __init__(self, pipeline_context):
        super(CalibrationMaker, self).__init__(pipeline_context)

    @abc.abstractproperty
    def calibration_type(self):
        pass

    @abc.abstractmethod
    def make_master_calibration_frame(self, images, image_config, logging_tags):
        pass

    @abc.abstractproperty
    def min_images(self):
        pass

    def do_stage(self, images):
        if len(images) < self.min_images:
            # Do nothing
            self.logger.warning('Not enough images to combine.')
            return []
        else:
            image_config = check_image_homogeneity(images)
            logging_tags = logs.image_config_to_tags(image_config, self.group_by_keywords)

            return self.make_master_calibration_frame(images, image_config, logging_tags)

    def get_calibration_filename(self, image):
        cal_file = '{cal_type}_{instrument}_{epoch}_bin{bin}{filter}.fits'
        if 'filter' in self.group_by_keywords:
            filter_str = '_{filter}'.format(filter=image.filter)
        else:
            filter_str = ''

        cal_file = cal_file.format(instrument=image.instrument,
                                   epoch=image.epoch, bin=image.ccdsum.replace(' ', 'x'),
                                   cal_type=self.calibration_type.lower(), filter=filter_str)
        return cal_file


class MasterCalibrationDoesNotExist(Exception):
    pass


class ApplyCalibration(Stage):
    __metaclass__ = abc.ABCMeta

    def __init__(self, pipeline_context):
        super(ApplyCalibration, self).__init__(pipeline_context)

    @abc.abstractproperty
    def calibration_type(self):
        pass

    def do_stage(self, images):
        if len(images) == 0:
            # Abort!
            return []
        else:
            image_config = check_image_homogeneity(images)
            logging_tags = logs.image_config_to_tags(image_config, self.group_by_keywords)
            master_calibration_filename = self.get_calibration_filename(images[0])

            if master_calibration_filename is None:
                self.logger.error('Master Calibration file does not exist for {stage}'.format(stage=self.stage_name),
                                  extra=logging_tags)
                raise MasterCalibrationDoesNotExist

            master_calibration_image = Image(master_calibration_filename)
            return self.apply_master_calibration(images, master_calibration_image, logging_tags)

    @abc.abstractmethod
    def apply_master_calibration(self, images, master_calibration_image, logging_tags):
        pass

    def get_calibration_filename(self, image):
        calibration_criteria = dbs.CalibrationImage.type == self.calibration_type.upper()
        calibration_criteria &= dbs.CalibrationImage.telescope_id == image.telescope_id

        for criterion in self.group_by_keywords:
            if criterion == 'filter':
                calibration_criteria &= dbs.CalibrationImage.filter_name == getattr(image, criterion)
            else:
                calibration_criteria &= getattr(dbs.CalibrationImage, criterion) == getattr(image, criterion)

        db_session = dbs.get_session()

        calibration_query = db_session.query(dbs.CalibrationImage).filter(calibration_criteria)
        epoch_datetime = date_utils.epoch_string_to_date(image.epoch)

        find_closest = func.DATEDIFF(epoch_datetime, dbs.CalibrationImage.dayobs)
        find_closest = func.ABS(find_closest)

        calibration_query = calibration_query.order_by(find_closest.asc())
        calibration_image = calibration_query.first()
        if calibration_image is None:
            calibration_file = None
        else:
            calibration_file = os.path.join(calibration_image.filepath, calibration_image.filename)

        db_session.close()
        return calibration_file



import banzai
from banzai import logs
from banzai.stages import Stage
from banzai.utils import date_utils
from datetime import timedelta
import os


class HeaderUpdater(Stage):
    def __init__(self, pipeline_context):
        super(HeaderUpdater, self).__init__(pipeline_context)

    @property
    def group_by_keywords(self):
        return None

    def do_stage(self, images):
        for image in images:
            image.header['RLEVEL'] = (self.pipeline_context.rlevel, 'Reduction level')
            image.header['PIPEVER'] = (banzai.__version__, 'Pipeline version')

            if instantly_public(image.header['PROPID']):
                image.header['L1PUBDAT'] = (image.header['DATE-OBS'],
                                            '[UTC] Date the frame becomes public')
            else:
                # Wait a year
                date_observed = date_utils.parse_date_obs(image.header['DATE-OBS'])
                next_year = date_observed + timedelta(days=365)
                image.header['L1PUBDAT'] = (date_utils.date_obs_to_string(next_year),
                                            '[UTC] Date the frame becomes public')
            logging_tags = logs.image_config_to_tags(image, self.group_by_keywords)
            logs.add_tag(logging_tags, 'filename', os.path.basename(image.filename))
            logs.add_tag(logging_tags, 'rlevel', int(image.header['RLEVEL']))
            logs.add_tag(logging_tags, 'pipeline_version', image.header['PIPEVER'])
            logs.add_tag(logging_tags, 'l1pubdat', image.header['L1PUBDAT'])
            self.logger.info('Updating header', extra=logging_tags)

        return images


def instantly_public(proposal_id):
    public_now = False
    if proposal_id in ['calibrate', 'standard', 'pointing']:
        public_now = True
    if 'epo' in proposal_id.lower():
        public_now = True
    return public_now



from __future__ import absolute_import, print_function, division
from astropy.io import fits
from astropy.coordinates import SkyCoord
from astropy import units
import numpy as np
import os
import tempfile
from . import date_utils

__author__ = 'cmccully'


def sanitizeheader(header):
    # Remove the mandatory keywords from a header so it can be copied to a new
    # image.
    header = header.copy()

    # Let the new data decide what these values should be
    for i in ['SIMPLE', 'BITPIX', 'BSCALE', 'BZERO']:
        if i in header.keys():
            header.pop(i)

    return header


def create_master_calibration_header(images):
    header = fits.Header()
    for h in images[0].header.keys():
        header[h] = images[0].header[h]

    header = sanitizeheader(header)

    observation_dates = [image.dateobs for image in images]
    mean_dateobs = date_utils.mean_date(observation_dates)

    header['DATE-OBS'] = date_utils.date_obs_to_string(mean_dateobs)

    header.add_history("Images combined to create master calibration image:")
    for image in images:
        header.add_history(image.filename)
    return header


def split_slice(pixel_section):
    pixels = pixel_section.split(':')
    if int(pixels[1]) > int(pixels[0]):
        pixel_slice = slice(int(pixels[0]) - 1, int(pixels[1]), 1)
    else:
        if int(pixels[1]) == 1:
            pixel_slice = slice(int(pixels[0]) - 1, None, -1)
        else:
            pixel_slice = slice(int(pixels[0]) - 1, int(pixels[1]) - 2, -1)
    return pixel_slice


def parse_region_keyword(keyword_value):
    """
    Convert a header keyword of the form [x1:x2],[y1:y2] into index slices
    :param keyword_value: Header keyword string
    :return: x, y index slices
    """

    if keyword_value.lower() == 'unknown':
        pixel_slices = None
    elif keyword_value.lower() == 'n/a':
        pixel_slices = None
    else:
        # Strip off the brackets and split the coordinates
        pixel_sections = keyword_value[1:-1].split(',')
        x_slice = split_slice(pixel_sections[0])
        y_slice = split_slice(pixel_sections[1])
        pixel_slices = (y_slice, x_slice)
    return pixel_slices


def table_to_fits(table):
    """
    Convert an astropy table to a fits binary table HDU
    :param table: astropy table
    :return: fits BinTableHDU
    """
    columns = [fits.Column(name=col.upper(), format=fits_formats(table[col].dtype),
                           array=table[col]) for col in table.colnames]
    return fits.BinTableHDU.from_columns(columns)


def parse_ra_dec(header):
    try:
        coord = SkyCoord(header.get('RA'), header.get('DEC'), unit=(units.hourangle, units.degree))
        ra = coord.ra.deg
        dec = coord.dec.deg
    except ValueError:
        ra, dec = np.nan, np.nan
    return ra, dec


def open_image(filename):
    base_filename = os.path.basename(filename)

    with tempfile.TemporaryDirectory() as tmpdirname:
        if filename[-3:] == '.fz':
            # Strip off the .fz
            output_filename = os.path.join(tmpdirname, base_filename)[:-3]
            os.system('funpack -O {0} {1}'.format(output_filename, filename))
            fits_filename = output_filename
        else:
            fits_filename = filename

        hdu = fits.open(fits_filename, 'readonly')
        data = hdu[0].data.astype(np.float32)
        header = hdu[0].header
        try:
            bpm = hdu['BPM'].data.astype(np.uint8)
        except KeyError:
            bpm = None
        hdu.close()

    return data, header, bpm


def fits_formats(dtype):
    """		
    Convert a numpy data type to a fits format code		
    :param dtype: dtype parameter from numpy array
    :return: string: Fits type code		
    """		
    format_code = ''		
    if 'bool' in dtype.name:
        format_code = 'L'		
    elif np.issubdtype(dtype, np.int16):
        format_code = 'I'		
    elif np.issubdtype(dtype, np.int32):
        format_code = 'J'		
    elif np.issubdtype(dtype, np.int64):
        format_code = 'K'		
    elif np.issubdtype(dtype, np.float32):
        format_code = 'E'		
    elif np.issubdtype(dtype, np.float64):
        format_code = 'D'		
    elif np.issubdtype(dtype, np.complex32):
        format_code = 'C'		
    elif np.issubdtype(dtype, np.complex64):
        format_code = 'M'		
    elif np.issubdtype(dtype, np.character):
        format_code = 'A'		
    return format_code		



""" main.py: Main driver script for banzai.

    The main() function is a console entry point.

Author
    Curtis McCully (cmccully@lcogt.net)

October 2015
"""
from __future__ import absolute_import, print_function, division

import argparse

from banzai import munge, crosstalk, gain, mosaic
from banzai import bias, dark, flats, trim, photometry, astrometry, headers
from banzai import logs
from banzai.utils import file_utils
from banzai import dbs
import os
import sys
from kombu.mixins import ConsumerMixin
from kombu import Connection, Queue, Exchange

# A dictionary converting the string input by the user into the corresponding Python object
reduction_stages = [bias.BiasMaker]

logger = logs.get_logger(__name__)


class PipelineContext(object):
    def __init__(self, args):
        self.processed_path = args.processed_path
        self.raw_path = args.raw_path
        self.post_to_archive = args.post_to_archive
        self.fpack = args.fpack
        self.rlevel = args.rlevel
        self.db_address = args.db_address
        self.log_level = args.log_level
        self.preview_mode = args.preview_mode
        self.filename = args.filename


def make_master_bias(cmd_args=None):
    pipeline_context = parse_command_line_arguments(cmd_args=cmd_args)
    logs.start_logging(log_level=pipeline_context.log_level)
    stages_to_do = [munge.DataMunger, crosstalk.CrosstalkCorrector, bias.OverscanSubtractor,
                    gain.GainNormalizer, mosaic.MosaicCreator, trim.Trimmer, bias.BiasMaker,
                    headers.HeaderUpdater]
    run(stages_to_do, pipeline_context, image_types=['BIAS'], calibration_maker=True,
        log_message='Making Master BIAS')
    logs.stop_logging()


def make_master_dark(cmd_args=None):
    pipeline_context = parse_command_line_arguments(cmd_args=cmd_args)
    logs.start_logging(log_level=pipeline_context.log_level)
    stages_to_do = [munge.DataMunger, crosstalk.CrosstalkCorrector, bias.OverscanSubtractor,
                    gain.GainNormalizer, mosaic.MosaicCreator, trim.Trimmer,
                    bias.BiasSubtractor, dark.DarkMaker, headers.HeaderUpdater]
    run(stages_to_do, pipeline_context, image_types=['DARK'], calibration_maker=True,
        log_message='Making Master Dark')
    logs.stop_logging()


def make_master_flat(cmd_args=None):
    pipeline_context = parse_command_line_arguments(cmd_args=cmd_args)
    logs.start_logging(log_level=pipeline_context.log_level)
    stages_to_do = [munge.DataMunger, crosstalk.CrosstalkCorrector, bias.OverscanSubtractor,
                    gain.GainNormalizer, mosaic.MosaicCreator, trim.Trimmer, bias.BiasSubtractor,
                    dark.DarkSubtractor, flats.FlatMaker, headers.HeaderUpdater]
    run(stages_to_do, pipeline_context, image_types=['SKYFLAT'], calibration_maker=True,
        log_message='Making Master Flat')
    logs.stop_logging()


def reduce_science_frames(cmd_args=None):

    pipeline_context = parse_command_line_arguments(cmd_args=cmd_args)
    logs.start_logging(log_level=pipeline_context.log_level)
    stages_to_do = [munge.DataMunger, crosstalk.CrosstalkCorrector, bias.OverscanSubtractor,
                    gain.GainNormalizer, mosaic.MosaicCreator, trim.Trimmer, bias.BiasSubtractor,
                    dark.DarkSubtractor, flats.FlatDivider, photometry.SourceDetector,
                    astrometry.WCSSolver, headers.HeaderUpdater]

    image_list = file_utils.make_image_list(pipeline_context)
    for image in image_list:
        pipeline_context.filename = os.path.basename(image)
        run(stages_to_do, pipeline_context, image_types=['EXPOSE', 'STANDARD'],
            log_message='Reducing Science Frames')
    logs.stop_logging()


def create_master_calibrations(cmd_args=None):
    make_master_bias(cmd_args=cmd_args)
    make_master_dark(cmd_args=cmd_args)
    make_master_flat(cmd_args=cmd_args)

    
def reduce_night(cmd_args=None):
    make_master_bias(cmd_args=cmd_args)
    make_master_dark(cmd_args=cmd_args)
    make_master_flat(cmd_args=cmd_args)
    reduce_science_frames(cmd_args=cmd_args)


def parse_command_line_arguments(cmd_args=None):
    parser = argparse.ArgumentParser(
        description='Make master calibration frames from LCOGT imaging data.')
    parser.add_argument("--raw-path", default='/archive/engineering',
                        help='Top level directory where the raw data is stored')
    parser.add_argument("--processed-path", default='/nethome/supernova/banzai',
                        help='Top level directory where the processed data will be stored')
    parser.add_argument("--log-level", default='info', choices=['debug', 'info', 'warning',
                                                                'critical', 'fatal', 'error'])
    parser.add_argument('--post-to-archive', dest='post_to_archive', action='store_true',
                        default=False)
    parser.add_argument('--db-address', dest='db_address',
                        default='mysql://cmccully:password@localhost/test',
                        help='Database address: Should be in SQLAlchemy form')
    parser.add_argument('--fpack', dest='fpack', action='store_true', default=False,
                        help='Fpack the output files?')
    parser.add_argument('--rlevel', dest='rlevel', default=91, help='Reduction level')
    parser.add_argument('--preview-mode', dest='preview_mode', action='store_true',
                        help='Store the data preview mode?')

    parser.add_argument('--filename', dest='filename', default=None,
                        help='Filename of the image to reduce.')
    args = parser.parse_args(cmd_args)

    return PipelineContext(args)


def run(stages_to_do, pipeline_context, image_types=[], calibration_maker=False, log_message=''):
    """
    Main driver script for banzai.
    """
    logger.info(log_message)

    image_list = file_utils.make_image_list(pipeline_context)
    image_list = file_utils.select_images(image_list, image_types)
    images = file_utils.read_images(image_list, pipeline_context)

    for stage in stages_to_do:
        stage_to_run = stage(pipeline_context)
        images = stage_to_run.run(images)

    file_utils.save_images(pipeline_context, images, master_calibration=calibration_maker)


def run_preview_pipeline(cmd_args=None):
    pipeline_context = parse_command_line_arguments(cmd_args=cmd_args)
    logs.start_logging(log_level=pipeline_context.log_level)
    logger.info('Starting pipeline preview mode listener')
    crawl_exchange = Exchange('fits_files', type='fanout')

    listener = PreviewModeListener('amqp://guest:guest@cerberus.lco.gtn', pipeline_context)

    with Connection(listener.broker_url) as connection:
        listener.connection = connection
        listener.queue = Queue('preview_pipeline', crawl_exchange)
        try:
            listener.run()
        except KeyboardInterrupt:
            logger.info('Shutting down preview pipeline listener...')
            logs.stop_logging()
            sys.exit(0)


class PreviewModeListener(ConsumerMixin):
    def __init__(self, broker_url, pipeline_context):
        self.broker_url = broker_url
        self.pipeline_context = pipeline_context

    def on_connection_error(self, exc, interval):
        logger.error("{0}. Retrying connection in {1} seconds...".format(exc, interval))

    def get_consumers(self, Consumer, channel):
        consumer = Consumer(queues=[self.queue], callbacks=[self.on_message])
        # Only fetch one thing off the queue at a time
        consumer.qos(prefetch_count=1)
        return [consumer]

    def on_message(self, body, message):
        path = body.get('path')
        if 'e00.fits' in path or 's00.fits' in path:
            if not dbs.preview_file_already_processed(path, db_address=self.pipeline_context.db_address):
                stages_to_do = [munge.DataMunger, crosstalk.CrosstalkCorrector,
                                bias.OverscanSubtractor, gain.GainNormalizer, mosaic.MosaicCreator,
                                trim.Trimmer, bias.BiasSubtractor, dark.DarkSubtractor,
                                flats.FlatDivider, photometry.SourceDetector, astrometry.WCSSolver,
                                headers.HeaderUpdater]
                logger.info('Running preview reduction on {}'.format(path))
                self.pipeline_context.filename = os.path.basename(path)
                self.pipeline_context.raw_path = os.path.dirname(path)
                run(stages_to_do, self.pipeline_context, image_types=['EXPOSE', 'STANDARD'])
                dbs.set_preview_file_as_processed(path, db_address=self.pipeline_context.db_address)

        message.ack()  # acknowledge to the sender we got this message (it can be popped)



#!/usr/bin/env python

"""
Setuptools bootstrapping installer.

Maintained at https://github.com/pypa/setuptools/tree/bootstrap.

Run this script to install or upgrade setuptools.
"""

import os
import shutil
import sys
import tempfile
import zipfile
import optparse
import subprocess
import platform
import textwrap
import contextlib
import json
import codecs

from distutils import log

try:
    from urllib.request import urlopen
    from urllib.parse import urljoin
except ImportError:
    from urllib2 import urlopen
    from urlparse import urljoin

try:
    from site import USER_SITE
except ImportError:
    USER_SITE = None

LATEST = object()
DEFAULT_VERSION = LATEST
DEFAULT_URL = "https://pypi.io/packages/source/s/setuptools/"
DEFAULT_SAVE_DIR = os.curdir


def _python_cmd(*args):
    """
    Execute a command.

    Return True if the command succeeded.
    """
    args = (sys.executable,) + args
    return subprocess.call(args) == 0


def _install(archive_filename, install_args=()):
    """Install Setuptools."""
    with archive_context(archive_filename):
        # installing
        log.warn('Installing Setuptools')
        if not _python_cmd('setup.py', 'install', *install_args):
            log.warn('Something went wrong during the installation.')
            log.warn('See the error message above.')
            # exitcode will be 2
            return 2


def _build_egg(egg, archive_filename, to_dir):
    """Build Setuptools egg."""
    with archive_context(archive_filename):
        # building an egg
        log.warn('Building a Setuptools egg in %s', to_dir)
        _python_cmd('setup.py', '-q', 'bdist_egg', '--dist-dir', to_dir)
    # returning the result
    log.warn(egg)
    if not os.path.exists(egg):
        raise IOError('Could not build the egg.')


class ContextualZipFile(zipfile.ZipFile):

    """Supplement ZipFile class to support context manager for Python 2.6."""

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def __new__(cls, *args, **kwargs):
        """Construct a ZipFile or ContextualZipFile as appropriate."""
        if hasattr(zipfile.ZipFile, '__exit__'):
            return zipfile.ZipFile(*args, **kwargs)
        return super(ContextualZipFile, cls).__new__(cls)


@contextlib.contextmanager
def archive_context(filename):
    """
    Unzip filename to a temporary directory, set to the cwd.

    The unzipped target is cleaned up after.
    """
    tmpdir = tempfile.mkdtemp()
    log.warn('Extracting in %s', tmpdir)
    old_wd = os.getcwd()
    try:
        os.chdir(tmpdir)
        with ContextualZipFile(filename) as archive:
            archive.extractall()

        # going in the directory
        subdir = os.path.join(tmpdir, os.listdir(tmpdir)[0])
        os.chdir(subdir)
        log.warn('Now working in %s', subdir)
        yield

    finally:
        os.chdir(old_wd)
        shutil.rmtree(tmpdir)


def _do_download(version, download_base, to_dir, download_delay):
    """Download Setuptools."""
    egg = os.path.join(to_dir, 'setuptools-%s-py%d.%d.egg'
                       % (version, sys.version_info[0], sys.version_info[1]))
    if not os.path.exists(egg):
        archive = download_setuptools(version, download_base,
                                      to_dir, download_delay)
        _build_egg(egg, archive, to_dir)
    sys.path.insert(0, egg)

    # Remove previously-imported pkg_resources if present (see
    # https://bitbucket.org/pypa/setuptools/pull-request/7/ for details).
    if 'pkg_resources' in sys.modules:
        _unload_pkg_resources()

    import setuptools
    setuptools.bootstrap_install_from = egg


def use_setuptools(
        version=DEFAULT_VERSION, download_base=DEFAULT_URL,
        to_dir=DEFAULT_SAVE_DIR, download_delay=15):
    """
    Ensure that a setuptools version is installed.

    Return None. Raise SystemExit if the requested version
    or later cannot be installed.
    """
    version = _resolve_version(version)
    to_dir = os.path.abspath(to_dir)

    # prior to importing, capture the module state for
    # representative modules.
    rep_modules = 'pkg_resources', 'setuptools'
    imported = set(sys.modules).intersection(rep_modules)

    try:
        import pkg_resources
        pkg_resources.require("setuptools>=" + version)
        # a suitable version is already installed
        return
    except ImportError:
        # pkg_resources not available; setuptools is not installed; download
        pass
    except pkg_resources.DistributionNotFound:
        # no version of setuptools was found; allow download
        pass
    except pkg_resources.VersionConflict as VC_err:
        if imported:
            _conflict_bail(VC_err, version)

        # otherwise, unload pkg_resources to allow the downloaded version to
        #  take precedence.
        del pkg_resources
        _unload_pkg_resources()

    return _do_download(version, download_base, to_dir, download_delay)


def _conflict_bail(VC_err, version):
    """
    Setuptools was imported prior to invocation, so it is
    unsafe to unload it. Bail out.
    """
    conflict_tmpl = textwrap.dedent("""
        The required version of setuptools (>={version}) is not available,
        and can't be installed while this script is running. Please
        install a more recent version first, using
        'easy_install -U setuptools'.

        (Currently using {VC_err.args[0]!r})
        """)
    msg = conflict_tmpl.format(**locals())
    sys.stderr.write(msg)
    sys.exit(2)


def _unload_pkg_resources():
    sys.meta_path = [
        importer
        for importer in sys.meta_path
        if importer.__class__.__module__ != 'pkg_resources.extern'
    ]
    del_modules = [
        name for name in sys.modules
        if name.startswith('pkg_resources')
    ]
    for mod_name in del_modules:
        del sys.modules[mod_name]


def _clean_check(cmd, target):
    """
    Run the command to download target.

    If the command fails, clean up before re-raising the error.
    """
    try:
        subprocess.check_call(cmd)
    except subprocess.CalledProcessError:
        if os.access(target, os.F_OK):
            os.unlink(target)
        raise


def download_file_powershell(url, target):
    """
    Download the file at url to target using Powershell.

    Powershell will validate trust.
    Raise an exception if the command cannot complete.
    """
    target = os.path.abspath(target)
    ps_cmd = (
        "[System.Net.WebRequest]::DefaultWebProxy.Credentials = "
        "[System.Net.CredentialCache]::DefaultCredentials; "
        '(new-object System.Net.WebClient).DownloadFile("%(url)s", "%(target)s")'
        % locals()
    )
    cmd = [
        'powershell',
        '-Command',
        ps_cmd,
    ]
    _clean_check(cmd, target)


def has_powershell():
    """Determine if Powershell is available."""
    if platform.system() != 'Windows':
        return False
    cmd = ['powershell', '-Command', 'echo test']
    with open(os.path.devnull, 'wb') as devnull:
        try:
            subprocess.check_call(cmd, stdout=devnull, stderr=devnull)
        except Exception:
            return False
    return True
download_file_powershell.viable = has_powershell


def download_file_curl(url, target):
    cmd = ['curl', url, '--location', '--silent', '--output', target]
    _clean_check(cmd, target)


def has_curl():
    cmd = ['curl', '--version']
    with open(os.path.devnull, 'wb') as devnull:
        try:
            subprocess.check_call(cmd, stdout=devnull, stderr=devnull)
        except Exception:
            return False
    return True
download_file_curl.viable = has_curl


def download_file_wget(url, target):
    cmd = ['wget', url, '--quiet', '--output-document', target]
    _clean_check(cmd, target)


def has_wget():
    cmd = ['wget', '--version']
    with open(os.path.devnull, 'wb') as devnull:
        try:
            subprocess.check_call(cmd, stdout=devnull, stderr=devnull)
        except Exception:
            return False
    return True
download_file_wget.viable = has_wget


def download_file_insecure(url, target):
    """Use Python to download the file, without connection authentication."""
    src = urlopen(url)
    try:
        # Read all the data in one block.
        data = src.read()
    finally:
        src.close()

    # Write all the data in one block to avoid creating a partial file.
    with open(target, "wb") as dst:
        dst.write(data)
download_file_insecure.viable = lambda: True


def get_best_downloader():
    downloaders = (
        download_file_powershell,
        download_file_curl,
        download_file_wget,
        download_file_insecure,
    )
    viable_downloaders = (dl for dl in downloaders if dl.viable())
    return next(viable_downloaders, None)


def download_setuptools(
        version=DEFAULT_VERSION, download_base=DEFAULT_URL,
        to_dir=DEFAULT_SAVE_DIR, delay=15,
        downloader_factory=get_best_downloader):
    """
    Download setuptools from a specified location and return its filename.

    `version` should be a valid setuptools version number that is available
    as an sdist for download under the `download_base` URL (which should end
    with a '/'). `to_dir` is the directory where the egg will be downloaded.
    `delay` is the number of seconds to pause before an actual download
    attempt.

    ``downloader_factory`` should be a function taking no arguments and
    returning a function for downloading a URL to a target.
    """
    version = _resolve_version(version)
    # making sure we use the absolute path
    to_dir = os.path.abspath(to_dir)
    zip_name = "setuptools-%s.zip" % version
    url = download_base + zip_name
    saveto = os.path.join(to_dir, zip_name)
    if not os.path.exists(saveto):  # Avoid repeated downloads
        log.warn("Downloading %s", url)
        downloader = downloader_factory()
        downloader(url, saveto)
    return os.path.realpath(saveto)


def _resolve_version(version):
    """
    Resolve LATEST version
    """
    if version is not LATEST:
        return version

    meta_url = urljoin(DEFAULT_URL, '/pypi/setuptools/json')
    resp = urlopen(meta_url)
    with contextlib.closing(resp):
        try:
            charset = resp.info().get_content_charset()
        except Exception:
            # Python 2 compat; assume UTF-8
            charset = 'UTF-8'
        reader = codecs.getreader(charset)
        doc = json.load(reader(resp))

    return str(doc['info']['version'])


def _build_install_args(options):
    """
    Build the arguments to 'python setup.py install' on the setuptools package.

    Returns list of command line arguments.
    """
    return ['--user'] if options.user_install else []


def _parse_args():
    """Parse the command line for options."""
    parser = optparse.OptionParser()
    parser.add_option(
        '--user', dest='user_install', action='store_true', default=False,
        help='install in user site package')
    parser.add_option(
        '--download-base', dest='download_base', metavar="URL",
        default=DEFAULT_URL,
        help='alternative URL from where to download the setuptools package')
    parser.add_option(
        '--insecure', dest='downloader_factory', action='store_const',
        const=lambda: download_file_insecure, default=get_best_downloader,
        help='Use internal, non-validating downloader'
    )
    parser.add_option(
        '--version', help="Specify which version to download",
        default=DEFAULT_VERSION,
    )
    parser.add_option(
        '--to-dir',
        help="Directory to save (and re-use) package",
        default=DEFAULT_SAVE_DIR,
    )
    options, args = parser.parse_args()
    # positional arguments are ignored
    return options


def _download_args(options):
    """Return args for download_setuptools function from cmdline args."""
    return dict(
        version=options.version,
        download_base=options.download_base,
        downloader_factory=options.downloader_factory,
        to_dir=options.to_dir,
    )


def main():
    """Install or upgrade setuptools and EasyInstall."""
    options = _parse_args()
    archive = download_setuptools(**_download_args(options))
    return _install(archive, _build_install_args(options))

if __name__ == '__main__':
    sys.exit(main())



from astropy.io import fits
from banzai.utils import date_utils
from banzai.utils import fits_utils
from banzai import dbs
import numpy as np
import os


class Image(object):

    def __init__(self, filename=None, data=None, header={}, bpm=None):

        if filename is not None:
            data, header, bpm = fits_utils.open_image(filename)
            if '.fz' == filename[-3:]:
                filename = filename[:-3]
            self.filename = os.path.basename(filename)

        self.data = data
        self.header = header
        self.bpm = bpm

        self.site = header.get('SITEID')
        self.instrument = header.get('INSTRUME')
        self.epoch = header.get('DAY-OBS')
        self.nx = header.get('NAXIS1')
        self.ny = header.get('NAXIS2')

        self.gain = header.get('GAIN')
        self.ccdsum = header.get('CCDSUM')
        self.filter = header.get('FILTER')
        self.telescope_id = dbs.get_telescope_id(self.site, self.instrument)

        self.obstype = header.get('OBSTYPE')
        self.exptime = float(header.get('EXPTIME'))
        self.dateobs = date_utils.parse_date_obs(header.get('DATE-OBS'))
        self.readnoise = float(header.get('RDNOISE'))
        self.ra, self.dec = fits_utils.parse_ra_dec(header)
        self.pixel_scale = float(header.get('PIXSCALE'))
        self.catalog = None

    def subtract(self, value):
        self.data -= value

    def writeto(self, filename, fpack=False):
        image_hdu = fits.PrimaryHDU(self.data.astype(np.float32), header=self.header)
        image_hdu.header['EXTEND'] = True
        image_hdu.update_ext_name('SCI')
        hdu_list = [image_hdu]
        if self.catalog is not None:
            table_hdu = fits_utils.table_to_fits(self.catalog)
            table_hdu.update_ext_name('CAT')
            hdu_list.append(table_hdu)
        if self.bpm is not None:
            bpm_hdu = fits.ImageHDU(self.bpm.astype(np.uint8))
            bpm_hdu.update_ext_name('BPM')
            hdu_list.append(bpm_hdu)

        hdu_list = fits.HDUList(hdu_list)
        hdu_list.writeto(filename, clobber=True)
        if fpack:
            if os.path.exists(filename + '.fz'):
                os.remove(filename + '.fz')
            os.system('fpack -q 64 {0}'.format(filename))
            os.remove(filename)
            self.filename += '.fz'

    def update_shape(self, nx, ny):
        self.nx = nx
        self.ny = ny

    def write_catalog(self, filename, nsources=None):
        if self.catalog is None:
            raise MissingCatalogException
        else:
            self.catalog[:nsources].write(filename, format='fits', overwrite=True)

    def add_history(self, msg):
        self.header.add_history(msg)


class InhomogeneousSetException(Exception):
    pass


def check_image_homogeneity(images):
    for attribute in ('nx', 'ny', 'ccdsum', 'epoch', 'site', 'instrument'):
        if len({getattr(image, attribute) for image in images}) > 1:
            raise InhomogeneousSetException('Images have different {}s'.format(attribute))
    return images[0]


class MissingCatalogException(Exception):
    pass



import errno
import os
import re
import shlex
import shutil
import subprocess
import sys
import textwrap

from distutils import log, ccompiler, sysconfig
from distutils.cmd import Command
from distutils.core import Extension
from distutils.ccompiler import get_default_compiler
from setuptools.command.build_ext import build_ext as SetuptoolsBuildExt

from ..utils import get_numpy_include_path, invalidate_caches, classproperty
from ..version_helpers import get_pkg_version_module


def should_build_with_cython(package, release=None):
    """Returns the previously used Cython version (or 'unknown' if not
    previously built) if Cython should be used to build extension modules from
    pyx files.  If the ``release`` parameter is not specified an attempt is
    made to determine the release flag from `astropy.version`.
    """

    try:
        version_module = __import__(package + '.cython_version',
                                    fromlist=['release', 'cython_version'])
    except ImportError:
        version_module = None

    if release is None and version_module is not None:
        try:
            release = version_module.release
        except AttributeError:
            pass

    try:
        cython_version = version_module.cython_version
    except AttributeError:
        cython_version = 'unknown'

    # Only build with Cython if, of course, Cython is installed, we're in a
    # development version (i.e. not release) or the Cython-generated source
    # files haven't been created yet (cython_version == 'unknown'). The latter
    # case can happen even when release is True if checking out a release tag
    # from the repository
    have_cython = False
    try:
        import Cython
        have_cython = True
    except ImportError:
        pass

    if have_cython and (not release or cython_version == 'unknown'):
        return cython_version
    else:
        return False


_compiler_versions = {}
def get_compiler_version(compiler):
    if compiler in _compiler_versions:
        return _compiler_versions[compiler]

    # Different flags to try to get the compiler version
    # TODO: It might be worth making this configurable to support
    # arbitrary odd compilers; though all bets may be off in such
    # cases anyway
    flags = ['--version', '--Version', '-version', '-Version',
             '-v', '-V']

    def try_get_version(flag):
        process = subprocess.Popen(
            shlex.split(compiler, posix=('win' not in sys.platform)) + [flag],
            stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        stdout, stderr = process.communicate()

        if process.returncode != 0:
            return 'unknown'

        output = stdout.strip().decode('latin-1')  # Safest bet
        if not output:
            # Some compilers return their version info on stderr
            output = stderr.strip().decode('latin-1')

        if not output:
            output = 'unknown'

        return output

    for flag in flags:
        version = try_get_version(flag)
        if version != 'unknown':
            break

    # Cache results to speed up future calls
    _compiler_versions[compiler] = version

    return version


# TODO: I think this can be reworked without having to create the class
# programmatically.
def generate_build_ext_command(packagename, release):
    """
    Creates a custom 'build_ext' command that allows for manipulating some of
    the C extension options at build time.  We use a function to build the
    class since the base class for build_ext may be different depending on
    certain build-time parameters (for example, we may use Cython's build_ext
    instead of the default version in distutils).

    Uses the default distutils.command.build_ext by default.
    """

    class build_ext(SetuptoolsBuildExt, object):
        package_name = packagename
        is_release = release
        _user_options = SetuptoolsBuildExt.user_options[:]
        _boolean_options = SetuptoolsBuildExt.boolean_options[:]
        _help_options = SetuptoolsBuildExt.help_options[:]

        force_rebuild = False


        _broken_compiler_mapping = [
            ('i686-apple-darwin[0-9]*-llvm-gcc-4.2', 'clang')
            ]

        # Warning: Spaghetti code ahead.
        # During setup.py, the setup_helpers module needs the ability to add
        # items to a command's user_options list.  At this stage we don't know
        # whether or not we can build with Cython, and so don't know for sure
        # what base class will be used for build_ext; nevertheless we want to
        # be able to provide a list to add options into.
        #
        # Later, once setup() has been called we should have all build
        # dependencies included via setup_requires available.  distutils needs
        # to be able to access the user_options as a *class* attribute before
        # the class has been initialized, but we do need to be able to
        # enumerate the options for the correct base class at that point

        @classproperty
        def user_options(cls):
            from distutils import core

            if core._setup_distribution is None:
                # We haven't gotten into setup() yet, and the Distribution has
                # not yet been initialized
                return cls._user_options

            return cls._final_class.user_options

        @classproperty
        def boolean_options(cls):
            # Similar to user_options above
            from distutils import core

            if core._setup_distribution is None:
                # We haven't gotten into setup() yet, and the Distribution has
                # not yet been initialized
                return cls._boolean_options

            return cls._final_class.boolean_options

        @classproperty
        def help_options(cls):
            # Similar to user_options above
            from distutils import core

            if core._setup_distribution is None:
                # We haven't gotten into setup() yet, and the Distribution has
                # not yet been initialized
                return cls._help_options

            return cls._final_class.help_options

        @classproperty(lazy=True)
        def _final_class(cls):
            """
            Late determination of what the build_ext base class should be,
            depending on whether or not Cython is available.
            """

            uses_cython = should_build_with_cython(cls.package_name,
                                                   cls.is_release)

            if uses_cython:
                # We need to decide late on whether or not to use Cython's
                # build_ext (since Cython may not be available earlier in the
                # setup.py if it was brought in via setup_requires)
                from Cython.Distutils import build_ext as base_cls
            else:
                base_cls = SetuptoolsBuildExt

            # Create and return an instance of a new class based on this class
            # using one of the above possible base classes
            def merge_options(attr):
                base = getattr(base_cls, attr)
                ours = getattr(cls, '_' + attr)

                all_base = set(opt[0] for opt in base)

                return base + [opt for opt in ours if opt[0] not in all_base]

            boolean_options = (base_cls.boolean_options +
                               [opt for opt in cls._boolean_options
                                if opt not in base_cls.boolean_options])

            members = dict(cls.__dict__)
            members.update({
                'user_options': merge_options('user_options'),
                'help_options': merge_options('help_options'),
                'boolean_options': boolean_options,
                'uses_cython': uses_cython,
            })

            # Update the base class for the original build_ext command
            build_ext.__bases__ = (base_cls, object)

            # Create a new class for the existing class, but now with the
            # appropriate base class depending on whether or not to use Cython.
            # Ensure that object is one of the bases to make a new-style class.
            return type(cls.__name__, (build_ext,), members)

        def __new__(cls, *args, **kwargs):
            # By the time the command is actually instantialized, the
            # Distribution instance for the build has been instantiated, which
            # means setup_requires has been processed--now we can determine
            # what base class we can use for the actual build, and return an
            # instance of a build_ext command that uses that base class (right
            # now the options being Cython.Distutils.build_ext, or the stock
            # setuptools build_ext)
            new_cls = super(build_ext, cls._final_class).__new__(
                    cls._final_class)

            # Since the new cls is not a subclass of the original cls, we must
            # manually call its __init__
            new_cls.__init__(*args, **kwargs)
            return new_cls

        def finalize_options(self):
            # Add a copy of the _compiler.so module as well, but only if there
            # are in fact C modules to compile (otherwise there's no reason to
            # include a record of the compiler used)
            # Note, self.extensions may not be set yet, but
            # self.distribution.ext_modules is where any extension modules
            # passed to setup() can be found
            self._adjust_compiler()

            extensions = self.distribution.ext_modules
            if extensions:
                src_path = os.path.relpath(
                    os.path.join(os.path.dirname(__file__), 'src'))
                shutil.copy2(os.path.join(src_path, 'compiler.c'),
                             os.path.join(self.package_name, '_compiler.c'))
                ext = Extension(self.package_name + '._compiler',
                                [os.path.join(self.package_name, '_compiler.c')])
                extensions.insert(0, ext)

            super(build_ext, self).finalize_options()

            # Generate
            if self.uses_cython:
                try:
                    from Cython import __version__ as cython_version
                except ImportError:
                    # This shouldn't happen if we made it this far
                    cython_version = None

                if (cython_version is not None and
                        cython_version != self.uses_cython):
                    self.force_rebuild = True
                    # Update the used cython version
                    self.uses_cython = cython_version

            # Regardless of the value of the '--force' option, force a rebuild
            # if the debug flag changed from the last build
            if self.force_rebuild:
                self.force = True

        def run(self):
            # For extensions that require 'numpy' in their include dirs,
            # replace 'numpy' with the actual paths
            np_include = get_numpy_include_path()
            for extension in self.extensions:
                if 'numpy' in extension.include_dirs:
                    idx = extension.include_dirs.index('numpy')
                    extension.include_dirs.insert(idx, np_include)
                    extension.include_dirs.remove('numpy')

                self._check_cython_sources(extension)

            super(build_ext, self).run()

            # Update cython_version.py if building with Cython
            try:
                cython_version = get_pkg_version_module(
                        packagename, fromlist=['cython_version'])[0]
            except (AttributeError, ImportError):
                cython_version = 'unknown'
            if self.uses_cython and self.uses_cython != cython_version:
                package_dir = os.path.relpath(packagename)
                cython_py = os.path.join(package_dir, 'cython_version.py')
                with open(cython_py, 'w') as f:
                    f.write('# Generated file; do not modify\n')
                    f.write('cython_version = {0!r}\n'.format(self.uses_cython))

                if os.path.isdir(self.build_lib):
                    # The build/lib directory may not exist if the build_py
                    # command was not previously run, which may sometimes be
                    # the case
                    self.copy_file(cython_py,
                                   os.path.join(self.build_lib, cython_py),
                                   preserve_mode=False)

                invalidate_caches()

        def _adjust_compiler(self):
            """
            This function detects broken compilers and switches to another.  If
            the environment variable CC is explicitly set, or a compiler is
            specified on the commandline, no override is performed -- the
            purpose here is to only override a default compiler.

            The specific compilers with problems are:

                * The default compiler in XCode-4.2, llvm-gcc-4.2,
                  segfaults when compiling wcslib.

            The set of broken compilers can be updated by changing the
            compiler_mapping variable.  It is a list of 2-tuples where the
            first in the pair is a regular expression matching the version of
            the broken compiler, and the second is the compiler to change to.
            """
            if 'CC' in os.environ:
                # Check that CC is not set to llvm-gcc-4.2
                c_compiler = os.environ['CC']

                try:
                    version = get_compiler_version(c_compiler)
                except OSError:
                    msg = textwrap.dedent(
                        """
                        The C compiler set by the CC environment variable:

                            {compiler:s}

                        cannot be found or executed.
                        """.format(compiler=c_compiler))
                    log.warn(msg)
                    sys.exit(1)

                for broken, fixed in self._broken_compiler_mapping:
                    if re.match(broken, version):
                        msg = textwrap.dedent(
                            """Compiler specified by CC environment variable
                            ({compiler:s}:{version:s}) will fail to compile
                            {pkg:s}.

                            Please set CC={fixed:s} and try again.
                            You can do this, for example, by running:

                                CC={fixed:s} python setup.py <command>

                            where <command> is the command you ran.
                            """.format(compiler=c_compiler, version=version,
                                       pkg=self.package_name, fixed=fixed))
                        log.warn(msg)
                        sys.exit(1)

                # If C compiler is set via CC, and isn't broken, we are good to go. We
                # should definitely not try accessing the compiler specified by
                # ``sysconfig.get_config_var('CC')`` lower down, because this may fail
                # if the compiler used to compile Python is missing (and maybe this is
                # why the user is setting CC). For example, the official Python 2.7.3
                # MacOS X binary was compiled with gcc-4.2, which is no longer available
                # in XCode 4.
                return

            if self.compiler is not None:
                # At this point, self.compiler will be set only if a compiler
                # was specified in the command-line or via setup.cfg, in which
                # case we don't do anything
                return

            compiler_type = ccompiler.get_default_compiler()

            if compiler_type == 'unix':
                # We have to get the compiler this way, as this is the one that is
                # used if os.environ['CC'] is not set. It is actually read in from
                # the Python Makefile. Note that this is not necessarily the same
                # compiler as returned by ccompiler.new_compiler()
                c_compiler = sysconfig.get_config_var('CC')

                try:
                    version = get_compiler_version(c_compiler)
                except OSError:
                    msg = textwrap.dedent(
                        """
                        The C compiler used to compile Python {compiler:s}, and
                        which is normally used to compile C extensions, is not
                        available. You can explicitly specify which compiler to
                        use by setting the CC environment variable, for example:

                            CC=gcc python setup.py <command>

                        or if you are using MacOS X, you can try:

                            CC=clang python setup.py <command>
                        """.format(compiler=c_compiler))
                    log.warn(msg)
                    sys.exit(1)

                for broken, fixed in self._broken_compiler_mapping:
                    if re.match(broken, version):
                        os.environ['CC'] = fixed
                        break

        def _check_cython_sources(self, extension):
            """
            Where relevant, make sure that the .c files associated with .pyx
            modules are present (if building without Cython installed).
            """

            # Determine the compiler we'll be using
            if self.compiler is None:
                compiler = get_default_compiler()
            else:
                compiler = self.compiler

            # Replace .pyx with C-equivalents, unless c files are missing
            for jdx, src in enumerate(extension.sources):
                base, ext = os.path.splitext(src)
                pyxfn = base + '.pyx'
                cfn = base + '.c'
                cppfn = base + '.cpp'


                if not os.path.isfile(pyxfn):
                    continue

                if self.uses_cython:
                    extension.sources[jdx] = pyxfn
                else:
                    if os.path.isfile(cfn):
                        extension.sources[jdx] = cfn
                    elif os.path.isfile(cppfn):
                        extension.sources[jdx] = cppfn
                    else:
                        msg = (
                            'Could not find C/C++ file {0}.(c/cpp) for Cython '
                            'file {1} when building extension {2}. Cython '
                            'must be installed to build from a git '
                            'checkout.'.format(base, pyxfn, extension.name))
                        raise IOError(errno.ENOENT, msg, cfn)

                # Current versions of Cython use deprecated Numpy API features
                # the use of which produces a few warnings when compiling.
                # These additional flags should squelch those warnings.
                # TODO: Feel free to remove this if/when a Cython update
                # removes use of the deprecated Numpy API
                if compiler == 'unix':
                    extension.extra_compile_args.extend([
                        '-Wp,-w', '-Wno-unused-function'])

    return build_ext



from __future__ import absolute_import, print_function, division
import numpy as np

from banzai.utils import median_utils

__author__ = 'cmccully'


def median(d, axis=None, mask=None):
    """
    Find the median of a numpy array. If an axis is provided, then find the median along
    the given axis. If mask is included, elements in d that have a non-zero mask value will
    be ignored.

    Parameters
    ----------
    d : float32 numpy array
        Input array to find the median.

    axis : int (default is None)
           Index of the array to take the median
    mask : unit8 or boolean numpy array (default is None)
          Numpy array of bitmask values. Non-zero values are ignored when calculating the median.

    Returns
    -------
    med : float32 numpy array
        The median value. If axis is None, then we return a single float.

    Notes
    -----
    Makes extensive use of the quick select algorithm written in C, included in quick_select.c.
    If all of the elements in the array are masked (or all of the elements of the axis of interest
    are masked), we return zero.
    """
    if axis is None:
        if mask is not None:
            median_mask = mask.ravel()
        else:
            median_mask = np.zeros(d.size, dtype=np.uint8)
        output_median = median_utils.median1d(np.ascontiguousarray(d.ravel(), dtype=np.float32),
                                              np.ascontiguousarray(median_mask, dtype=np.uint8))
    else:

        nx = d.shape[axis]
        ny = d.size // nx

        output_shape = np.delete(d.shape, axis)

        if mask is not None:
            median_mask = np.rollaxis(mask, axis, len(d.shape)).reshape(ny, nx).astype(np.uint8)
        else:
            median_mask = np.zeros((ny, nx), dtype=np.uint8)

        med = median_utils.median2d(np.ascontiguousarray(np.rollaxis(d, axis, len(d.shape)).reshape(ny, nx), dtype=np.float32),
                                    mask=np.ascontiguousarray(median_mask, dtype=np.uint8))
        median_array = np.array(med)
        output_median = median_array.reshape(output_shape)

    return output_median


def absolute_deviation(a, axis=None, mask=None):
    """
    Find the absolute deviation from the median of a numpy array. If an axis is provided,
    then the median is calculated along the given axis.
    If mask is included, elements in a that have a non-zero mask value will be ignored.

    Parameters
    ----------
    a : float32 numpy array
        Input array to find the absolute deviation from the median.

    axis : int (default is None)
           Index of the array to take the median
    mask : unit8 or boolean numpy array (default is None)
          Numpy array of bitmask values. Non-zero values are ignored when calculating the median.

    Returns
    -------
    absdev : float32 numpy array
        The absolute deviation from the median. If axis is None, then we return a single float.

    Notes
    -----
    If all of the elements in the array are masked (or all of the elements of the axis of interest
    are masked), the original array is returned.
    """

    a_median = median(a, axis=axis, mask=mask)
    if axis is not None:
        a_median = np.expand_dims(a_median, axis=axis)
    return np.abs(a - a_median)


def median_absolute_deviation(a, axis=None, abs_deviation=None, mask=None):
    if abs_deviation is None:
        abs_deviation = absolute_deviation(a, axis=axis, mask=mask)

    return median(abs_deviation, axis=axis, mask=mask)


def robust_standard_deviation(a, axis=None, abs_deviation=None, mask=None):
    return 1.4826 * median_absolute_deviation(a, axis=axis, abs_deviation=abs_deviation, mask=mask)


def sigma_clipped_mean(a, sigma, axis=None, mask=None, fill_value=0.0, inplace=False):
    """
    """
    abs_deviation = absolute_deviation(a, axis=axis, mask=mask)

    robust_std = robust_standard_deviation(a, axis=axis, abs_deviation=abs_deviation, mask=mask)

    if axis is not None:
        robust_std = np.expand_dims(robust_std, axis=axis)

    # Throw away any values that are N sigma from the median
    sigma_mask = abs_deviation > (sigma * robust_std)

    if mask is not None:
        sigma_mask = np.logical_or(sigma_mask, mask > 0)
    if inplace:
        mean_array = a
    else:
        mean_array = a.copy()

    mean_array[sigma_mask] = 0.0

    # Take the sigma clipped mean
    mean_values = mean_array.sum(axis=axis)

    n_good_pixels = np.logical_not(sigma_mask).sum(axis=axis)
    if axis is None:
        if n_good_pixels > 0:
            mean_values /= n_good_pixels
        else:
            mean_values = fill_value
    else:
        mean_values[n_good_pixels > 0] /= n_good_pixels[n_good_pixels > 0]
        mean_values[n_good_pixels == 0] = fill_value

    return mean_values



# Licensed under a 3-clause BSD style license - see LICENSE.rst

# This sub-module is destined for common non-package specific utility
# functions that will ultimately be merged into `astropy.utils`

from banzai.utils import *
__all__ = ['stats']



from banzai.stages import Stage
import numpy as np
from banzai.utils import fits_utils
from banzai import logs
import os


class MosaicCreator(Stage):
    def __init__(self, pipeline_context):
        super(MosaicCreator, self).__init__(pipeline_context)

    @property
    def group_by_keywords(self):
        return None

    def do_stage(self, images):
        for image in images:
            if len(image.data.shape) > 2:

                logging_tags = logs.image_config_to_tags(image, self.group_by_keywords)
                logs.add_tag(logging_tags, 'filename', os.path.basename(image.filename))

                n_amps = image.data.shape[0]
                nx, ny = get_mosaic_size(image, n_amps)
                mosaiced_data = np.zeros((ny, nx), dtype=np.float32)
                mosaiced_bpm = np.zeros((ny, nx), dtype=np.uint8)
                for i in range(n_amps):
                    datasec = image.header['DATASEC{0}'.format(i + 1)]
                    amp_slice = fits_utils.parse_region_keyword(datasec)
                    logs.add_tag(logging_tags, 'DATASEC{0}'.format(i + 1), datasec)

                    detsec = image.header['DETSEC{0}'.format(i + 1)]
                    mosaic_slice = fits_utils.parse_region_keyword(detsec)
                    logs.add_tag(logging_tags, 'DETSEC{0}'.format(i + 1), datasec)

                    mosaiced_data[mosaic_slice] = image.data[i][amp_slice]
                    mosaiced_bpm[mosaic_slice] = image.bpm[i][amp_slice]

                image.data = mosaiced_data
                image.bpm = mosaiced_bpm
                image.update_shape(nx, ny)
                image.header['NAXIS'] = 2
                image.header.pop('NAXIS3')

                self.logger.info('Mosaiced image', extra=logging_tags)
        return images


def get_mosaic_size(image, n_amps):
    x_pixel_limits = []
    y_pixel_limits = []
    for i in range(n_amps):
        header_keyword = image.header['DETSEC{0}'.format(i + 1)]
        y_slice, x_slice = fits_utils.parse_region_keyword(header_keyword)
        x_pixel_limits.append(x_slice.start + 1)
        x_pixel_limits.append(x_slice.stop)
        y_pixel_limits.append(y_slice.start + 1)
        y_pixel_limits.append(y_slice.stop)

    # Clean out any Nones
    x_pixel_limits = [x if x is not None else 1 for x in x_pixel_limits]
    y_pixel_limits = [y if y is not None else 1 for y in y_pixel_limits]

    nx = np.max(x_pixel_limits) - np.min(x_pixel_limits) + 1
    ny = np.max(y_pixel_limits) - np.min(y_pixel_limits) + 1
    return nx, ny



from __future__ import absolute_import, print_function, division
import os
import numpy as np

from astropy.table import Table
from banzai.utils import stats

from banzai.stages import Stage
from banzai import logs

import sep

__author__ = 'cmccully'


class SourceDetector(Stage):
    # Note that threshold is number of sigma, not an absolute number because we provide the error
    # array to SEP.
    threshold = 10.0
    min_area = 9

    def __init__(self, pipeline_context):
        super(SourceDetector, self).__init__(pipeline_context)

    @property
    def group_by_keywords(self):
        return None

    def do_stage(self, images):
        for i, image in enumerate(images):
            try:
                # Set the number of source pixels to be 5% of the total. This keeps us safe from
                # satellites and airplanes.
                sep.set_extract_pixstack(int(image.nx * image.ny * 0.05))

                data = image.data.copy()
                error = (np.abs(data) + image.readnoise ** 2.0) ** 0.5
                mask = image.bpm > 0

                # Fits can be backwards byte order, so fix that if need be and subtract
                # the background
                try:
                    bkg = sep.Background(data, mask=mask, bw=32, bh=32, fw=3, fh=3)
                except ValueError:
                    data = data.byteswap(True).newbyteorder()
                    bkg = sep.Background(data, mask=mask, bw=32, bh=32, fw=3, fh=3)
                bkg.subfrom(data)

                # Do an initial source detection
                # TODO: Add back in masking after we are sure SEP works
                sources = sep.extract(data, self.threshold, minarea=self.min_area,
                                      err=error, deblend_cont=0.005)

                # Convert the detections into a table
                sources = Table(sources)

                # Calculate the ellipticity
                sources['ellipticity'] = 1.0 - (sources['b'] / sources['a'])

                # Fix any value of theta that are invalid due to floating point rounding
                # -pi / 2 < theta < pi / 2
                sources['theta'][sources['theta'] > (np.pi / 2.0)] -= np.pi
                sources['theta'][sources['theta'] < (-np.pi / 2.0)] += np.pi

                # Calculate the kron radius
                kronrad, krflag = sep.kron_radius(data, sources['x'], sources['y'],
                                                  sources['a'], sources['b'],
                                                  sources['theta'], 6.0)
                sources['flag'] |= krflag
                sources['kronrad'] = kronrad

                # Calcuate the equivilent of flux_auto
                flux, fluxerr, flag = sep.sum_ellipse(data, sources['x'], sources['y'],
                                                      sources['a'], sources['b'],
                                                      np.pi / 2.0, 2.5 * kronrad,
                                                      subpix=1, err=error)
                sources['flux'] = flux
                sources['fluxerr'] = fluxerr
                sources['flag'] |= flag

                # Calculate the FWHMs of the stars:
                fwhm = 2.0 * (np.log(2) * (sources['a'] ** 2.0 + sources['b'] ** 2.0)) ** 0.5
                sources['fwhm'] = fwhm

                # Cut individual bright pixels. Often cosmic rays
                sources = sources[fwhm > 1.0]

                # Measure the flux profile
                flux_radii, flag = sep.flux_radius(data, sources['x'], sources['y'],
                                                   6.0 * sources['a'], [0.25, 0.5, 0.75],
                                                   normflux=sources['flux'], subpix=5)
                sources['flag'] |= flag
                sources['fluxrad25'] = flux_radii[:, 0]
                sources['fluxrad50'] = flux_radii[:, 1]
                sources['fluxrad75'] = flux_radii[:, 2]

                # Calculate the windowed positions
                sig = 2.0 / 2.35 * sources['fluxrad50']
                xwin, ywin, flag = sep.winpos(data, sources['x'], sources['y'], sig)
                sources['flag'] |= flag
                sources['xwin'] = xwin
                sources['ywin'] = ywin

                # Calculate the average background at each source
                bkgflux, fluxerr, flag = sep.sum_ellipse(bkg.back(), sources['x'], sources['y'],
                                                         sources['a'], sources['b'], np.pi / 2.0,
                                                         2.5 * sources['kronrad'], subpix=1)
                #masksum, fluxerr, flag = sep.sum_ellipse(mask, sources['x'], sources['y'],
                #                                         sources['a'], sources['b'], np.pi / 2.0,
                #                                         2.5 * kronrad, subpix=1)

                background_area = (2.5 * sources['kronrad']) ** 2.0 * sources['a'] * sources['b'] * np.pi # - masksum
                sources['background'] = bkgflux
                sources['background'][background_area > 0] /= background_area[background_area > 0]
                # Update the catalog to match fits convention instead of python array convention
                sources['x'] += 1.0
                sources['y'] += 1.0

                sources['xpeak'] += 1
                sources['ypeak'] += 1

                sources['xwin'] += 1.0
                sources['ywin'] += 1.0

                sources['theta'] = np.degrees(sources['theta'])

                image.catalog = sources['x', 'y', 'xwin', 'ywin', 'xpeak', 'ypeak',
                                        'flux', 'fluxerr', 'background', 'fwhm',
                                        'a', 'b', 'theta', 'kronrad', 'ellipticity',
                                        'fluxrad25', 'fluxrad50', 'fluxrad75',
                                        'x2', 'y2', 'xy', 'flag']

                # Add the units and description to the catalogs
                image.catalog['x'].unit = 'pixel'
                image.catalog['x'].description = 'X coordinate of the object'
                image.catalog['y'].unit = 'pixel'
                image.catalog['y'].description = 'Y coordinate of the object'
                image.catalog['xwin'].unit = 'pixel'
                image.catalog['xwin'].description = 'Windowed X coordinate of the object'
                image.catalog['ywin'].unit = 'pixel'
                image.catalog['ywin'].description = 'Windowed Y coordinate of the object'
                image.catalog['xpeak'].unit = 'pixel'
                image.catalog['xpeak'].description = 'X coordinate of the peak'
                image.catalog['ypeak'].unit = 'pixel'
                image.catalog['ypeak'].description = 'Windowed Y coordinate of the peak'
                image.catalog['flux'].unit = 'counts'
                image.catalog['flux'].description = 'Flux within a Kron-like elliptical aperture'
                image.catalog['fluxerr'].unit = 'counts'
                image.catalog['fluxerr'].description = 'Erronr on the flux within a Kron-like elliptical aperture'
                image.catalog['background'].unit = 'counts'
                image.catalog['background'].description = 'Average background value in the aperture'
                image.catalog['fwhm'].unit = 'pixel'
                image.catalog['fwhm'].description = 'FWHM of the object'
                image.catalog['a'].unit = 'pixel'
                image.catalog['a'].description = 'Semi-major axis of the object'
                image.catalog['b'].unit = 'pixel'
                image.catalog['b'].description = 'Semi-minor axis of the object'
                image.catalog['theta'].unit = 'degrees'
                image.catalog['theta'].description = 'Position angle of the object'
                image.catalog['kronrad'].unit = 'pixel'
                image.catalog['kronrad'].description = 'Kron radius used for extraction'
                image.catalog['ellipticity'].description = 'Ellipticity'
                image.catalog['fluxrad25'].unit = 'pixel'
                image.catalog['fluxrad25'].description = 'Radius containing 25% of the flux'
                image.catalog['fluxrad50'].unit = 'pixel'
                image.catalog['fluxrad50'].description = 'Radius containing 50% of the flux'
                image.catalog['fluxrad75'].unit = 'pixel'
                image.catalog['fluxrad75'].description = 'Radius containing 75% of the flux'
                image.catalog['x2'].unit = 'pixel^2'
                image.catalog['x2'].description = 'Variance on X coordinate of the object'
                image.catalog['y2'].unit = 'pixel^2'
                image.catalog['y2'].description = 'Variance on Y coordinate of the object'
                image.catalog['xy'].unit = 'pixel^2'
                image.catalog['xy'].description = 'XY covariance of the object'
                image.catalog['flag'].description = 'Bit mask combination of extraction and photometry flags'

                image.catalog.sort('flux')
                image.catalog.reverse()

                logging_tags = logs.image_config_to_tags(image, self.group_by_keywords)
                logs.add_tag(logging_tags, 'filename', os.path.basename(image.filename))

                # Save some background statistics in the header
                mean_background = stats.sigma_clipped_mean(bkg.back(), 5.0)
                image.header['L1MEAN'] = (mean_background,
                                          '[counts] Sigma clipped mean of frame background')
                logs.add_tag(logging_tags, 'L1MEAN', float(mean_background))

                median_background = np.median(bkg.back())
                image.header['L1MEDIAN'] = (median_background,
                                            '[counts] Median of frame background')
                logs.add_tag(logging_tags, 'L1MEDIAN', float(median_background))

                std_background = stats.robust_standard_deviation(bkg.back())
                image.header['L1SIGMA'] = (std_background,
                                           '[counts] Robust std dev of frame background')
                logs.add_tag(logging_tags, 'L1SIGMA', float(std_background))

                # Save some image statistics to the header
                good_objects = image.catalog['flag'] == 0

                seeing = np.median(image.catalog['fwhm'][good_objects]) * image.pixel_scale
                image.header['L1FWHM'] = (seeing, '[arcsec] Frame FWHM in arcsec')
                logs.add_tag(logging_tags, 'L1FWHM', float(seeing))

                mean_ellipticity = stats.sigma_clipped_mean(sources['ellipticity'][good_objects],
                                                            3.0)
                image.header['L1ELLIP'] = (mean_ellipticity, 'Mean image ellipticity (1-B/A)')
                logs.add_tag(logging_tags, 'L1ELLIP', float(mean_ellipticity))

                mean_position_angle = stats.sigma_clipped_mean(sources['theta'][good_objects], 3.0)
                image.header['L1ELLIPA'] = (mean_position_angle,
                                            '[deg] PA of mean image ellipticity')
                logs.add_tag(logging_tags, 'L1ELLIPA', float(mean_position_angle))

                self.logger.info('Extracted sources', extra=logging_tags)

            except Exception as e:
                logging_tags = logs.image_config_to_tags(image, self.group_by_keywords)
                logs.add_tag(logging_tags, 'filename', os.path.basename(image.filename))
                self.logger.error(e, extra=logging_tags)
        return images



from banzai.stages import Stage
import numpy as np
import os
from banzai import dbs
from banzai.utils import array_utils


class DataMunger(Stage):
    def __init__(self, pipeline_context):
        super(DataMunger, self).__init__(pipeline_context)

    @property
    def group_by_keywords(self):
        return None

    def do_stage(self, images):
        for image in images:
            telescope = dbs.get_telescope(image.telescope_id)
            # TODO: Currently we only support 1x1 Sinistro frames and only support 1x1 frames.
            # TODO: 2x2 frames cannot use hard coded values because we read out more pixels.
            if 'sinistro' in telescope.camera_type.lower():
                keywords_to_update = [('BIASSEC1', ('[1:2048,2055:2080]',
                                                    '[binned pixel] Section of overscan data for Q1')),
                                      ('BIASSEC2', ('[1:2048,2055:2080]',
                                                    '[binned pixel] Section of overscan data for  Q2')),
                                      ('BIASSEC3', ('[1:2048,2055:2080]',
                                                    '[binned pixel] Section of overscan data for Q3')),
                                      ('BIASSEC4', ('[1:2048,2055:2080]',
                                                    '[binned pixel] Section of overscan data for Q4')),
                                      ('DATASEC1', ('[1:2048,1:2048]',
                                                    '[binned pixel] Data section for  Q1')),
                                      ('DATASEC2', ('[1:2048,1:2048]',
                                                    '[binned pixel] Data section for Q2'))]

                if image.data.shape[1] > 2048:
                    keywords_to_update.append(('DATASEC3',  ('[1:2048,2:2049]',
                                               '[binned pixel] Data section for Q3')))
                    keywords_to_update.append(('DATASEC4', ('[1:2048,2:2049]',
                                               '[binned pixel] Data section for Q4')))
                else:
                    keywords_to_update.append(('DATASEC3', ('[1:2048,1:2048]',
                                                            '[binned pixel] Data section for Q3')))
                    keywords_to_update.append(('DATASEC4', ('[1:2048,1:2048]',
                                                            '[binned pixel] Data section for Q4')))
                keywords_to_update.append(('DETSEC1', ('[1:2048,1:2048]',
                                                       '[binned pixel] Detector section for Q1')))
                keywords_to_update.append(('DETSEC2', ('[4096:2049,1:2048]',
                                                       '[binned pixel] Detector section for Q2')))
                keywords_to_update.append(('DETSEC3', ('[4096:2049,4096:2049]',
                                                       '[binned pixel] Detector section for Q3')))
                keywords_to_update.append(('DETSEC4', ('[1:2048,4096:2049]',
                                                       '[binned pixel] Detector section for Q4')))

                set_crosstalk_header_keywords(image)

                for keyword, value in keywords_to_update:
                    _add_header_keyword(keyword, value, image)
            # 1m SBIGS
            elif '1m0' in telescope.camera_type:
                image.header['SATURATE'] = (46000.0, '[ADU] Saturation level used')
            elif '0m4' in telescope.camera_type or '0m8' in telescope.camera_type:
                image.header['SATURATE'] = (56000.0, '[ADU] Saturation level used')
            elif 'fs02' == telescope.instrument:
                # These values were given by Joe Tufts on 2016-06-07
                # These should really be measured empirically.
                if image.header['CCDSUM'] == '2 2':
                    image.header['SATURATE'] = (500000.0 / float(image.header['GAIN']),
                                                '[ADU] Saturation level used')
                elif image.header['CCDSUM'] == '1 1':
                    image.header['SATURATE'] = (125000.0 / float(image.header['GAIN']),
                                                '[ADU] Saturation level used')
            # Throw an exception if the saturate value is set to zero
            try:
                if float(image.header['SATURATE']) == 0.0:
                    images.remove(image)
                    raise ValueError('SATURATE keyword cannot be zero: {filename}'.format(filename=os.path.basename(image.filename)))
            except ValueError:
                continue

            bpm_slices = array_utils.array_indices_to_slices(image.bpm)
            image.bpm[image.data[bpm_slices] >= float(image.header['SATURATE'])] = 2
        return images


def _add_header_keyword(keyword, value, image):
    if image.header.get(keyword) is None:
        image.header[keyword] = value


def set_crosstalk_header_keywords(image):
    n_amps = image.data.shape[0]
    coefficients = crosstalk_coefficients.get(image.instrument, np.zeros((n_amps, n_amps)))

    for j in range(n_amps):
        for i in range(n_amps):
            if i != j:
                crosstalk_comment = '[Crosstalk coefficient] Q{i} signal due to Q{j}'.format(i=i+1, j=j+1)
                image.header['CRSTLK{0}{1}'.format(i + 1, j + 1)] = (coefficients[j, i], crosstalk_comment)

crosstalk_coefficients = {'fl01': np.array([[0.00000, 0.00074, 0.00081, 0.00115],
                                            [0.00070, 0.00000, 0.00118, 0.00085],
                                            [0.00076, 0.00115, 0.00000, 0.00088],
                                            [0.00107, 0.00075, 0.00080, 0.00000]]),
                          'fl02': np.array([[0.00000, 0.00084, 0.00088, 0.00125],
                                            [0.00083, 0.00000, 0.00124, 0.00096],
                                            [0.00086, 0.00121, 0.00000, 0.00098],
                                            [0.00116, 0.00085, 0.00092, 0.00000]]),
                          'fl03': np.array([[0.00000, 0.00076, 0.00079, 0.00115],
                                            [0.00073, 0.00000, 0.00117, 0.00084],
                                            [0.00074, 0.00113, 0.00000, 0.00084],
                                            [0.00105, 0.00075, 0.00080, 0.00000]]),
                          'fl04': np.array([[0.00000, 0.00088, 0.00096, 0.00131],
                                            [0.00087, 0.00000, 0.00132, 0.00099],
                                            [0.00087, 0.00127, 0.00000, 0.00103],
                                            [0.00123, 0.00089, 0.00094, 0.00000]]),
                          'fl05': np.array([[0.00000, 0.00084, 0.00090, 0.00126],
                                            [0.00089, 0.00000, 0.00133, 0.00095],
                                            [0.00097, 0.00155, 0.00000, 0.00108],
                                            [0.00134, 0.00096, 0.00095, 0.00000]]),
                          'fl06': np.array([[0.00000, 0.00078, 0.00085, 0.00118],
                                            [0.00075, 0.00000, 0.00112, 0.00087],
                                            [0.00074, 0.00110, 0.00000, 0.00089],
                                            [0.00098, 0.00075, 0.00081, 0.00000]]),
                          'fl07': np.array([[0.00000, 0.00075, 0.00077, 0.00113],
                                            [0.00071, 0.00000, 0.00113, 0.00082],
                                            [0.00070, 0.00108, 0.00000, 0.00086],
                                            [0.00095, 0.00067, 0.00077, 0.00000]]),
                          'fl08': np.array([[0.00000, 0.00057, 0.00078, 0.00130],
                                            [0.00112, 0.00000, 0.00163, 0.00123],
                                            [0.00104, 0.00113, 0.00000, 0.00113],
                                            [0.00108, 0.00048, 0.00065, 0.00000]])
                          }



from __future__ import absolute_import, print_function, division

import logutils.queue
import logging
import multiprocessing
import sys
from lcogt_logging import LCOGTFormatter
from .utils import date_utils

def get_logger(name):
    logger = logging.getLogger(name)
    # This allows us to control the logging level with the root logger.
    logger.setLevel(logging.DEBUG)
    return logger


def start_logging(log_level='INFO', filename=None):

    # Set up the message queue
    global queue
    queue = multiprocessing.Queue(-1)

    # Start the listener process
    global listener
    listener = logutils.queue.QueueListener(queue)
    listener.start()

    # Set up the logging format
    root_logger = logging.getLogger()
    if filename is not None:
        root_handler = logging.FileHandler(filename)
    else:
        root_handler = logging.StreamHandler(sys.stdout)

    def get_process_name():
        return multiprocessing.current_process().name

    formatter = LCOGTFormatter(extra_tags={'processName': get_process_name})
    root_handler.setFormatter(formatter)
    root_handler.setLevel(getattr(logging, log_level.upper(), None))
    root_logger.addHandler(root_handler)

    queue_handler = logutils.queue.QueueHandler(queue)

    queue_handler.setLevel(logging.DEBUG)
    root_logger.addHandler(queue_handler)


def stop_logging():
    if listener is not None:
        queue.put_nowait(None)
        listener.stop()


def image_config_to_tags(image_config, group_by_keywords):
    tags = {'tags': {'site': image_config.site,
                     'instrument': image_config.instrument,
                     'epoch': date_utils.epoch_date_to_string(image_config.epoch)}}
    if group_by_keywords is not None:
        for i, keyword in enumerate(group_by_keywords):
            tags['tags'][keyword] = getattr(image_config, keyword)

    return tags


def add_tag(tags, keyword, value):
    tags['tags'][keyword] = value


def pop_tag(tags, keyword):
    tags['tags'].pop(keyword)


import pytest
from ..images import InhomogeneousSetException
import numpy as np
from datetime import datetime


class FakeImage(object):
    def __init__(self, nx=101, ny=103, image_multiplier=1.0,
                 ccdsum='2 2', epoch='20160101',):
        self.nx = nx
        self.ny = ny
        self.telescope_id = -1
        self.site = 'elp'
        self.instrument = 'kb76'
        self.ccdsum = ccdsum
        self.epoch = epoch
        self.data = image_multiplier * np.ones((ny, nx), dtype=np.float32)
        self.filename = 'test.fits'
        self.filter = 'U'
        self.dateobs = datetime(2016, 1, 1)
        self.header = {}
        self.caltype = ''
        self.bpm = np.zeros((ny, nx), dtype=np.uint8)

    def get_calibration_filename(self):
        return '/tmp/{0}_{1}_{2}_bin{3}.fits'.format(self.caltype, self.instrument,
                                                     self.epoch,
                                                     self.ccdsum.replace(' ', 'x'))

    def subtract(self, x):
        self.data -= x

    def add_history(self, msg):
        pass


class FakeContext(object):
    def __init__(self):
        self.processed_path = '/tmp'


def throws_inhomogeneous_set_exception(stagetype, context, keyword, value):
    stage = stagetype(context)

    with pytest.raises(InhomogeneousSetException) as exception_info:
        kwargs = {keyword: value}
        images = [FakeImage(**kwargs)]
        images += [FakeImage() for x in range(6)]
        stage.do_stage(images)
    assert 'Images have different {0}s'.format(keyword) == str(exception_info.value)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This packages contains affiliated package tests.
"""
from . import test_bias_maker, test_bias_subtractor
__all__ = ['test_bias_maker', 'test_bias_subtractor']



from __future__ import absolute_import, print_function, division

import numpy as np
import os.path

from .utils import stats, fits_utils
from .stages import CalibrationMaker, ApplyCalibration
from banzai.images import Image
from banzai import logs

__author__ = 'cmccully'


class FlatMaker(CalibrationMaker):

    def __init__(self, pipeline_context):

        super(FlatMaker, self).__init__(pipeline_context)

    @property
    def calibration_type(self):
        return 'skyflat'

    @property
    def group_by_keywords(self):
        return ['ccdsum', 'filter']

    @property
    def min_images(self):
        return 5

    def make_master_calibration_frame(self, images, image_config, logging_tags):
        flat_data = np.zeros((images[0].ny, images[0].nx, len(images)), dtype=np.float32)
        flat_mask = np.zeros((images[0].ny, images[0].nx, len(images)), dtype=np.uint8)

        quarter_nx = images[0].nx // 4
        quarter_ny = images[0].ny // 4

        master_flat_filename = self.get_calibration_filename(images[0])
        logs.add_tag(logging_tags, 'master_flat', os.path.basename(master_flat_filename))
        for i, image in enumerate(images):

            # Get the sigma clipped mean of the central 25% of the image
            flat_normalization = stats.sigma_clipped_mean(image.data[quarter_ny: -quarter_ny,
                                                                     quarter_nx:-quarter_nx], 3.5)
            flat_data[:, :, i] = image.data[:, :]
            flat_data[:, :, i] /= flat_normalization
            flat_mask[:, :, i] = image.bpm[:, :]
            logs.add_tag(logging_tags, 'filename', os.path.basename(image.filename))
            logs.add_tag(logging_tags, 'flat_normalization', flat_normalization)
            self.logger.debug('Calculating flat normalization', extra=logging_tags)

        logs.pop_tag(logging_tags, 'flat_normalization')
        master_flat = stats.sigma_clipped_mean(flat_data, 3.0, axis=2, mask=flat_mask,
                                               fill_value=1.0, inplace=True)

        master_bpm = np.array(master_flat == 1.0, dtype=np.uint8)

        master_bpm = np.logical_and(master_bpm, master_flat < 0.2)

        master_flat[master_flat < 0.2] = 1.0

        master_flat_header = fits_utils.create_master_calibration_header(images)

        master_flat_image = Image(data=master_flat, header=master_flat_header)
        master_flat_image.filename = master_flat_filename
        master_flat_image.bpm = master_bpm

        logs.pop_tag(logging_tags, 'master_flat')
        logs.add_tag(logging_tags, 'filename', os.path.basename(master_flat_image.filename))
        self.logger.info('Created master flat', extra=logging_tags)

        return [master_flat_image]


class FlatDivider(ApplyCalibration):
    def __init__(self, pipeline_context):

        super(FlatDivider, self).__init__(pipeline_context)

    @property
    def group_by_keywords(self):
        return ['ccdsum', 'filter']

    @property
    def calibration_type(self):
        return 'skyflat'

    def apply_master_calibration(self, images, master_calibration_image, logging_tags):

        master_flat_filename = master_calibration_image.filename
        master_flat_data = master_calibration_image.data
        logs.add_tag(logging_tags, 'master_flat',
                     os.path.basename(master_calibration_image.filename))
        for image in images:
            logs.add_tag(logging_tags, 'filename', os.path.basename(image.filename))
            self.logger.info('Flattening image', extra=logging_tags)
            image.data /= master_flat_data
            image.bpm |= master_calibration_image.bpm
            master_flat_filename = os.path.basename(master_flat_filename)
            image.header['L1IDFLAT'] = (master_flat_filename, 'ID of flat frame used')
            image.header['L1STATFL'] = (1, 'Status flag for flat field correction')

        return images



from banzai.stages import Stage
from banzai import logs
import numpy as np
import os


class CrosstalkCorrector(Stage):
    def __init__(self, pipeline_context):
        super(CrosstalkCorrector, self).__init__(pipeline_context)

    @property
    def group_by_keywords(self):
        return None

    def do_stage(self, images):
        for image in images:
            if len(image.data.shape) > 2:
                logging_tags = logs.image_config_to_tags(image, self.group_by_keywords)
                logs.add_tag(logging_tags, 'filename', os.path.basename(image.filename))
                n_amps = image.data.shape[0]
                crosstalk_matrix = np.identity(n_amps)
                for j in range(n_amps):
                    for i in range(n_amps):
                        if i != j:
                            crosstalk_keyword = 'CRSTLK{0}{1}'.format(i + 1, j + 1)
                            crosstalk_matrix[j, i] = -float(image.header[crosstalk_keyword])
                            logs.add_tag(logging_tags, crosstalk_keyword,
                                         image.header[crosstalk_keyword])
                self.logger.info('Removing crosstalk', extra=logging_tags)
                image.data = np.dot(crosstalk_matrix, np.swapaxes(image.data, 0, 1))
        return images



from ..utils import median_utils
import numpy as np

def _compare_median1d(a, mask):
    a_copy = a.copy()
    actual = median_utils.median1d(a.astype('f4'), mask=mask.astype(np.uint8))
    expected = np.median(a_copy[mask == 0].astype('f4'))
    assert actual == np.float32(expected)


def test_median1d_arange_nomask():
    for i in range(100):
        size = np.random.choice(10000)
        a = np.arange(size, dtype=np.float32)
        np.random.shuffle(a)
        _compare_median1d(a, np.zeros(size, dtype=np.uint8))


def test_median1d_normal_distribution_nomask():
    for i in range(100):
        mean = np.random.uniform(-1000.0, 1000.0)
        sigma = np.random.uniform(0, 100.0)
        size = np.random.choice(10000)
        a = np.random.normal(mean, sigma, size=size)
        _compare_median1d(a.astype('f4'), np.zeros(size, dtype=np.uint8))


def test_median1d_uniform_distribution_nomask():
    for i in range(100):
        size = np.random.choice(10000)
        a = np.random.uniform(-1000.0, 1000.0, size=size)
        _compare_median1d(a.astype('f4'), np.zeros(size, dtype=np.uint8))


def test_median1d_bimodel_arange_nomask():
    for i in range(100):
        size1 = np.random.randint(1, 10000)
        size2 = np.random.randint(1, 10000)
        start1 = np.random.randint(0, 10000)
        start2 = np.random.randint(0, 10000)
        a = np.append(np.arange(size1, dtype=np.float32) + start1,
                      np.arange(size2, dtype=np.float32) + start2)
        _compare_median1d(a, np.zeros(size1 + size2, dtype=np.uint8))


def test_median1d_bimodal_normal_distribution_nomask():
    for i in range(100):
        mean1 = np.random.uniform(-1000.0, 1000.0)
        sigma1 = np.random.uniform(0, 100.0)
        mean2 = np.random.uniform(-1000.0, 1000.0)
        sigma2 = np.random.uniform(0, 100.0)
        size1 = np.random.randint(1, 10000)
        size2 = np.random.randint(1, 10000)
        a = np.append(np.random.normal(mean1, sigma1, size=size1),
                      np.random.normal(mean2, sigma2, size=size2))
        _compare_median1d(a, np.zeros(size1 + size2, dtype=np.uint8))


def test_median1d_bimodal_uniformal_distribution_nomask():
    for i in range(100):
        size1 = np.random.randint(1, 10000)
        size2 = np.random.randint(1, 10000)
        center1 = np.random.uniform(-10000.0, 10000.0)
        center2 = np.random.uniform(-10000.0, 10000.0)
        a = np.append(np.random.normal(-1000.0, 1000.0, size=size1) + center1,
                      np.random.normal(-1000.0, 1000.0, size=size2) + center2)
        _compare_median1d(a, np.zeros(size1 + size2, dtype=np.uint8))


def test_median1d_all_masks_returns_zero():
    size = 1000
    a = np.zeros(size, dtype=np.float32)
    mask = np.ones(size, dtype=np.uint8)
    assert median_utils.median1d(a, mask) == 0.0


def test_median1d_arange_mask():
    for i in range(100):
        size = np.random.randint(1, 10000)
        a = np.arange(size, dtype=np.float32)
        np.random.shuffle(a)
        mask = np.zeros(size, dtype=np.uint8)
        mask[np.random.randint(0, size - 1, size=np.random.randint(1, size - 1))] = 1
        _compare_median1d(a, mask)


def test_median1d_normal_distribution_mask():
    for i in range(100):
        mean = np.random.uniform(-1000.0, 1000.0)
        sigma = np.random.uniform(0, 100.0)
        size = np.random.randint(1, 10000)
        a = np.random.normal(mean, sigma, size=size)
        mask = np.zeros(size, dtype=np.uint8)
        mask[np.random.randint(0, size - 1, size=np.random.randint(1, size - 1))] = 1
        _compare_median1d(a.astype('f4'), mask)


def test_median1d_uniform_distribution_mask():
    for i in range(100):
        size = np.random.randint(1, 10000)
        a = np.random.uniform(-1000.0, 1000.0, size=size)
        mask = np.zeros(size, dtype=np.uint8)
        mask[np.random.randint(0, size - 1, size=np.random.randint(1, size - 1))] = 1
        _compare_median1d(a.astype('f4'), mask)


def test_median1d_bimodel_arange_mask():
    for i in range(100):
        size1 = np.random.randint(1, 10000)
        size2 = np.random.randint(1, 10000)
        start1 = np.random.randint(0, 10000)
        start2 = np.random.randint(0, 10000)
        mask = np.zeros(size1 + size2, dtype=np.uint8)
        mask[np.random.randint(0, size1 + size2 - 1, size=np.random.randint(1, size1 + size2 - 1))] = 1
        a = np.append(np.arange(size1, dtype=np.float32) + start1,
                      np.arange(size2, dtype=np.float32) + start2)
        _compare_median1d(a, mask)


def test_median1d_bimodal_normal_distribution_mask():
    for i in range(100):
        mean1 = np.random.uniform(-1000.0, 1000.0)
        sigma1 = np.random.uniform(0, 100.0)
        mean2 = np.random.uniform(-1000.0, 1000.0)
        sigma2 = np.random.uniform(0, 100.0)
        size1 = np.random.randint(1, 10000)
        size2 = np.random.randint(1, 10000)
        a = np.append(np.random.normal(mean1, sigma1, size=size1),
                      np.random.normal(mean2, sigma2, size=size2))
        mask = np.zeros(size1 + size2, dtype=np.uint8)
        mask[np.random.randint(0, size1 + size2 - 1, size=np.random.randint(1, size1 + size2 - 1))] = 1
        _compare_median1d(a, mask)


def test_median1d_bimodal_uniformal_distribution_mask():
    for i in range(100):
        size1 = np.random.randint(1, 10000)
        size2 = np.random.randint(1, 10000)
        center1 = np.random.uniform(-10000.0, 10000.0)
        center2 = np.random.uniform(-10000.0, 10000.0)
        mask = np.zeros(size1 + size2, dtype=np.uint8)
        mask[np.random.randint(0, size1 + size2 - 1, size=np.random.randint(1, size1 + size2 - 1))] = 1
        a = np.append(np.random.normal(-1000.0, 1000.0, size=size1) + center1,
                      np.random.normal(-1000.0, 1000.0, size=size2) + center2)
        _compare_median1d(a, mask)


def _compare_median2d(a, mask):
    actual = median_utils.median2d(a.astype('f4'), mask.astype(np.uint8))

    mask_array = mask == 0
    n_good_pixels = mask_array.sum(axis=1)[0]
    d = a[mask == 0].reshape(a.shape[0], n_good_pixels)
    expected = np.median(d.astype('f4'), axis=1)
    np.testing.assert_allclose(actual, np.float32(expected))


def test_median2d_arange_nomask():
    for i in range(10):
        size_x, size_y = np.random.randint(1, 1000), np.random.randint(1, 1000)
        a = np.arange(size_x * size_y, dtype=np.float32).reshape(size_y, size_x)
        np.random.shuffle(a)
        _compare_median2d(a, np.zeros((size_y, size_x), dtype=np.uint8))


def test_median2d_normal_distribution_nomask():
    for i in range(10):
        mean = np.random.uniform(-1000.0, 1000.0)
        sigma = np.random.uniform(0, 100.0)
        size_x, size_y = np.random.randint(1, 1000), np.random.randint(1, 1000)
        a = np.random.normal(mean, sigma, size=(size_y, size_x))
        _compare_median2d(a.astype('f4'), np.zeros((size_y, size_x), dtype=np.uint8))


def test_median2d_uniform_distribution_nomask():
    for i in range(10):
        size_x, size_y = np.random.randint(1, 1000), np.random.randint(1, 1000)
        a = np.random.uniform(-1000.0, 1000.0, size=(size_y, size_x))
        _compare_median2d(a.astype('f4'), np.zeros((size_y, size_x), dtype=np.uint8))


# def test_median2d_bimodel_arange_nomask():
#     for i in range(100):
#         size1 = np.random.randint(1, 10000)
#         size2 = np.random.randint(1, 10000)
#         start1 = np.random.choice(10000)
#         start2 = np.random.choice(10000)
#         a = np.append(np.arange(size1, dtype=np.float32) + start1,
#                       np.arange(size2, dtype=np.float32) + start2)
#         _compare_median2d(a, np.zeros(size1 + size2, dtype=np.uint8))
#
#
# def test_median2d_bimodal_normal_distribution_nomask():
#     for i in range(100):
#         mean1 = np.random.uniform(-1000.0, 1000.0)
#         sigma1 = np.random.uniform(0, 100.0)
#         mean2 = np.random.uniform(-1000.0, 1000.0)
#         sigma2 = np.random.uniform(0, 100.0)
#         size1 = np.random.randint(1, 10000)
#         size2 = np.random.randint(1, 10000)
#         a = np.append(np.random.normal(mean1, sigma1, size=size1),
#                       np.random.normal(mean2, sigma2, size=size2))
#         _compare_median2d(a, np.zeros(size1 + size2, dtype=np.uint8))
#
#
# def test_median2d_bimodal_uniformal_distribution_nomask():
#     for i in range(100):
#         size1 = np.random.randint(1, 10000)
#         size2 = np.random.randint(1, 10000)
#         center1 = np.random.uniform(-10000.0, 10000.0)
#         center2 = np.random.uniform(-10000.0, 10000.0)
#         a = np.append(np.random.normal(-1000.0, 1000.0, size=size1) + center1,
#                       np.random.normal(-1000.0, 1000.0, size=size2) + center2)
#         _compare_median2d(a, np.zeros(size1 + size2, dtype=np.uint8))
#
#
# def test_median2d_all_masks_returns_zero():
#     size = 1000
#     a = np.zeros(size, dtype=np.float32)
#     mask = np.ones(size, dtype=np.uint8)
#     assert stats._median2d(a, mask) == 0.0
#
#
# def test_median2d_arange_mask():
#     for i in range(100):
#         size = np.random.choice(10000)
#         a = np.arange(size, dtype=np.float32)
#         np.random.shuffle(a)
#         mask = np.zeros(size, dtype=np.uint8)
#         mask[np.random.choice(size - 1, size=np.random.randint(1, size - 1))] = 1
#         _compare_median2d(a, mask)
#
#
# def test_median2d_normal_distribution_mask():
#     for i in range(100):
#         mean = np.random.uniform(-1000.0, 1000.0)
#         sigma = np.random.uniform(0, 100.0)
#         size = np.random.choice(10000)
#         a = np.random.normal(mean, sigma, size=size)
#         mask = np.zeros(size, dtype=np.uint8)
#         mask[np.random.choice(size - 1, size=np.random.randint(1, size - 1))] = 1
#         _compare_median2d(a.astype('f4'), mask)
#
#
# def test_median2d_uniform_distribution_mask():
#     for i in range(100):
#         size = np.random.choice(10000)
#         a = np.random.uniform(-1000.0, 1000.0, size=size)
#         mask = np.zeros(size, dtype=np.uint8)
#         mask[np.random.choice(size - 1, size=np.random.randint(1, size - 1))] = 1
#         _compare_median2d(a.astype('f4'), mask)
#
#
# def test_median2d_bimodel_arange_mask():
#     for i in range(100):
#         size1 = np.random.randint(1, 10000)
#         size2 = np.random.randint(1, 10000)
#         start1 = np.random.choice(10000)
#         start2 = np.random.choice(10000)
#         mask = np.zeros(size1 + size2, dtype=np.uint8)
#         mask[np.random.choice(size1 + size2 - 1, size=np.random.randint(1, size1 + size2 - 1))] = 1
#         a = np.append(np.arange(size1, dtype=np.float32) + start1,
#                       np.arange(size2, dtype=np.float32) + start2)
#         _compare_median2d(a, mask)
#
#
# def test_median2d_bimodal_normal_distribution_mask():
#     for i in range(100):
#         mean1 = np.random.uniform(-1000.0, 1000.0)
#         sigma1 = np.random.uniform(0, 100.0)
#         mean2 = np.random.uniform(-1000.0, 1000.0)
#         sigma2 = np.random.uniform(0, 100.0)
#         size1 = np.random.randint(1, 10000)
#         size2 = np.random.randint(1, 10000)
#         a = np.append(np.random.normal(mean1, sigma1, size=size1),
#                       np.random.normal(mean2, sigma2, size=size2))
#         mask = np.zeros(size1 + size2, dtype=np.uint8)
#         mask[np.random.choice(size1 + size2 - 1, size=np.random.randint(1, size1 + size2 - 1))] = 1
#         _compare_median2d(a, mask)
#
#
# def test_median2d_bimodal_uniformal_distribution_mask():
#     for i in range(100):
#         size1 = np.random.randint(1, 10000)
#         size2 = np.random.randint(1, 10000)
#         center1 = np.random.uniform(-10000.0, 10000.0)
#         center2 = np.random.uniform(-10000.0, 10000.0)
#         mask = np.zeros(size1 + size2, dtype=np.uint8)
#         mask[np.random.choice(size1 + size2 - 1, size=np.random.randint(1, size1 + size2 - 1))] = 1
#         a = np.append(np.random.normal(-1000.0, 1000.0, size=size1) + center1,
#                       np.random.normal(-1000.0, 1000.0, size=size2) + center2)
#         _compare_median2d(a, mask)
#


from ..utils import median_utils
import numpy as np


def test_quick_select_arange():
    for i in range(100):
        size = np.random.randint(2, 10000)
        index = np.random.randint(0, size - 1)
        expected = index
        a = np.arange(size, dtype=np.float32)
        np.random.shuffle(a)
        assert median_utils._quick_select(a, index) == np.float32(expected)


def _compare_quick_select(a, index):
    actual = median_utils._quick_select(a.astype('f4'), index)
    a = a.astype('f4')
    a.sort()
    expected = a[index]
    assert actual == np.float32(expected)


def test_quick_select_normal_distribution():
    for i in range(100):
        mean = np.random.uniform(-1000.0, 1000.0)
        sigma = np.random.uniform(0, 100.0)
        size = np.random.randint(2, 10000)
        index = np.random.randint(0, size - 1)
        a = np.random.normal(mean, sigma, size=size)
        _compare_quick_select(a, index)


def test_quick_select_uniform_distribution():
    for i in range(100):
        size = np.random.randint(2, 10000)
        index = np.random.randint(0, size - 2)
        a = np.random.uniform(-1000.0, 1000.0, size=size)
        _compare_quick_select(a, index)


def test_quick_select_bimodel_arange():
    for i in range(100):
        size1 = np.random.randint(2, 10000)
        size2 = np.random.randint(2, 10000)
        start1 = np.random.randint(0, 10000)
        start2 = np.random.randint(0, 10000)
        index = np.random.randint(0, size1 + size2 - 1)
        a = np.append(np.arange(size1, dtype=np.float32) + start1,
                      np.arange(size2, dtype=np.float32) + start2)
        np.random.shuffle(a)
        _compare_quick_select(a, index)


def test_quick_select_bimodal_normal_distribution():
    for i in range(100):
        mean1 = np.random.uniform(-1000.0, 1000.0)
        sigma1 = np.random.uniform(0, 100.0)
        mean2 = np.random.uniform(-1000.0, 1000.0)
        sigma2 = np.random.uniform(0, 100.0)
        size1 = np.random.randint(2, 10000)
        size2 = np.random.randint(2, 10000)
        index = np.random.randint(0, size1 + size2 - 1)
        a = np.append(np.random.normal(mean1, sigma1, size=size1),
                      np.random.normal(mean2, sigma2, size=size2))
        _compare_quick_select(a, index)


def test_quick_select_bimodal_uniformal_distribution():
    for i in range(100):
        size1 = np.random.randint(2, 10000)
        size2 = np.random.randint(2, 10000)
        center1 = np.random.uniform(-10000.0, 10000.0)
        center2 = np.random.uniform(-10000.0, 10000.0)
        index = np.random.randint(0, size1 + size2 - 1)
        a = np.append(np.random.normal(-1000.0, 1000.0, size=size1) + center1,
                      np.random.normal(-1000.0, 1000.0, size=size2) + center2)
        _compare_quick_select(a, index)



from __future__ import absolute_import, print_function, division
from banzai.stages import Stage
from banzai import logs
from banzai.images import MissingCatalogException
import os, subprocess, shlex
from astropy.io import fits
import tempfile
from astropy.wcs import WCS

__author__ = 'cmccully'


class WCSSolver(Stage):
    cmd = 'solve-field --crpix-center --no-verify --no-fits2fits --no-tweak ' \
          ' --radius 2.0 --ra {ra} --dec {dec} --guess-scale ' \
          '--scale-units arcsecperpix --scale-low {scale_low} --scale-high {scale_high} ' \
          '--no-plots -N none --use-sextractor --no-remove-lines ' \
          '--code-tolerance 0.003 --pixel-error 20 -d 1-200 ' \
          '--solved none --match none --rdls none --wcs {wcs_name} --corr none --overwrite ' \
          '-X X -Y Y -s FLUX --width {nx} --height {ny} {catalog_name}'

    def __init__(self, pipeline_context):
        super(WCSSolver, self).__init__(pipeline_context)

    @property
    def group_by_keywords(self):
        return None

    def do_stage(self, images):

        for i, image in enumerate(images):
            logging_tags = logs.image_config_to_tags(image, self.group_by_keywords)

            # Save the catalog to a temporary file
            filename = os.path.basename(image.filename)
            logs.add_tag(logging_tags, 'filename', filename)

            with tempfile.TemporaryDirectory() as tmpdirname:
                catalog_name = os.path.join(tmpdirname, filename.replace('.fits', '.cat.fits'))
                try:
                    image.write_catalog(catalog_name, nsources=40)
                except MissingCatalogException:
                    image.header['WCSERR'] = (4, 'Error status of WCS fit. 0 for no error')
                    self.logger.error('No source catalog. Not attempting WCS solution',
                                      extra=logging_tags)
                    continue

                # Run astrometry.net
                wcs_name = os.path.join(tmpdirname, filename.replace('.fits', '.wcs.fits'))
                command = self.cmd.format(ra=image.ra, dec=image.dec, scale_low=0.9*image.pixel_scale,
                                          scale_high=1.1*image.pixel_scale, wcs_name=wcs_name,
                                          catalog_name=catalog_name, nx=image.nx, ny=image.ny)

                console_output = subprocess.check_output(shlex.split(command))
                self.logger.debug(console_output, extra=logging_tags)

                if os.path.exists(wcs_name):
                    # Copy the WCS keywords into original image
                    new_header = fits.getheader(wcs_name)

                    header_keywords_to_update = ['CTYPE1', 'CTYPE2', 'CRPIX1', 'CRPIX2', 'CRVAL1',
                                                 'CRVAL2', 'CD1_1', 'CD1_2', 'CD2_1', 'CD2_2']
                    for keyword in header_keywords_to_update:
                        image.header[keyword] = new_header[keyword]

                    image.header['WCSERR'] = (0, 'Error status of WCS fit. 0 for no error')

                    # Clean up wcs file
                    os.remove(wcs_name)

                    # Add the RA and Dec values to the catalog
                    add_ra_dec_to_catalog(image)
                else:
                    image.header['WCSERR'] = (4, 'Error status of WCS fit. 0 for no error')

            logs.add_tag(logging_tags, 'WCSERR', image.header['WCSERR'])
            self.logger.info('Attempted WCS Solve', extra=logging_tags)
        return images


def add_ra_dec_to_catalog(image):
    image_wcs = WCS(image.header)
    ras, decs = image_wcs.all_pix2world(image.catalog['x'], image.catalog['y'], 1)
    image.catalog['ra'] = ras
    image.catalog['dec'] = decs
    image.catalog['ra'].unit = 'degrees'
    image.catalog['dec'].unit = 'degrees'
    image.catalog['ra'].description = 'Right Ascension'
    image.catalog['dec'].description = 'Declination'



"""

@author: mnorbury
"""
import os

import numpy as np

import pytest

from astropy.io import fits

from banzai.main import *
from banzai.dbs import create_db


def setup_function(function):
    create_db('sqlite:///test.db')


def teardown_function(function):
    os.remove('test.db')


def full_pipeline():
    data_path = '/nethome/cmccully/workspace/banzai/banzai/tests/data/'
    site = 'elp'
    instrument = 'kb74'
    epoch = '20150325'
    final_image_name = 'elp1m008-kb74-20150325-0123-e90.fits'

    main('--db-host sqlite:///test.db --raw-path {0:s} --instrument {1:s} --processed-path {0:s} '\
         '--log-level debug --site {2:s} --epoch {3:s} '.format(data_path, instrument, site, epoch).split())

    expected_file_path = os.path.join(data_path, final_image_name)
    expected_hdu = fits.open(expected_file_path)

    actual_file_path = os.path.join(data_path, site, instrument, epoch, final_image_name)
    actual_hdu = fits.open(actual_file_path)

    assert expected_hdu[0].header == actual_hdu[0].header
    assert np.allclose(expected_hdu[0].data, actual_hdu[0].data, 1e-6)


if __name__ == '__main__':
    pytest.main([__file__, '--cov=banzai', '--cov-report=html', '--pdb'])



from ..bias import BiasMaker
import numpy as np

from .utils import FakeImage, FakeContext, throws_inhomogeneous_set_exception

import mock


class FakeBiasImage(FakeImage):
    def __init__(self, *args, **kwargs):
        super(FakeBiasImage, self).__init__(*args, **kwargs)
        self.caltype = 'bias'
        self.header = {'OBSTYPE': 'BIAS'}


def test_min_images():

    bias_maker = BiasMaker(None)
    processed_images = bias_maker.do_stage([])
    assert len(processed_images) == 0


def test_group_by_keywords():
    maker = BiasMaker(None)
    assert maker.group_by_keywords == ['ccdsum']


@mock.patch('banzai.bias.Image')
def test_header_master_bias_level_returns_1(mock_image):
    maker = BiasMaker(FakeContext())

    maker.do_stage([FakeBiasImage() for x in range(6)])

    args, kwargs = mock_image.call_args
    header = kwargs['header']
    assert header['BIASLVL'] == 1.0


@mock.patch('banzai.bias.Image')
def test_header_master_bias_level_returns_2(mock_image):
    maker = BiasMaker(FakeContext())

    maker.do_stage([FakeBiasImage(image_multiplier=2.0) for x in range(6)])

    args, kwargs = mock_image.call_args
    header = kwargs['header']
    assert header['BIASLVL'] == 2.0


@mock.patch('banzai.bias.Image')
def test_header_cal_type_bias(mock_image):

    maker = BiasMaker(FakeContext())

    maker.do_stage([FakeBiasImage() for x in range(6)])

    args, kwargs = mock_image.call_args
    header = kwargs['header']
    assert header['OBSTYPE'].upper() == 'BIAS'


@mock.patch('banzai.bias.Image')
def test_raises_an_exection_if_ccdsums_are_different(mock_images):
    throws_inhomogeneous_set_exception(BiasMaker, FakeContext(), 'ccdsum', '1 1')


@mock.patch('banzai.bias.Image')
def test_raises_an_exection_if_epochs_are_different(mock_images):
    throws_inhomogeneous_set_exception(BiasMaker, FakeContext(), 'epoch', '20160102')


@mock.patch('banzai.bias.Image')
def test_raises_an_exection_if_nx_are_different(mock_images):
    throws_inhomogeneous_set_exception(BiasMaker, FakeContext(), 'nx', 105)


@mock.patch('banzai.bias.Image')
def test_raises_an_exection_if_ny_are_different(mock_images):
    throws_inhomogeneous_set_exception(BiasMaker, FakeContext(), 'ny', 107)


@mock.patch('banzai.bias.Image')
def test_makes_a_sensible_master_bias(mock_images):
    nimages = 20
    expected_bias = 1183.0
    expected_readnoise = 15.0

    images = [FakeBiasImage() for x in range(nimages)]
    for image in images:
        image.data = np.random.normal(loc=expected_bias, scale=expected_readnoise,
                                      size=(image.ny, image.nx))

    maker = BiasMaker(FakeContext())
    maker.do_stage(images)

    args, kwargs = mock_images.call_args
    master_bias = kwargs['data']
    assert np.abs(np.mean(master_bias)) < 0.1
    actual_bias = float(kwargs['header']['BIASLVL'])
    assert np.abs(actual_bias - expected_bias) < 0.1
    actual_readnoise = np.std(master_bias)
    assert np.abs(actual_readnoise - expected_readnoise / (nimages ** 0.5)) < 0.2



from __future__ import absolute_import, print_function, division

import os.path

import numpy as np

from banzai.images import Image
from . import logs
from .stages import CalibrationMaker, ApplyCalibration, Stage
from .utils import stats, fits_utils


__author__ = 'cmccully'


class BiasMaker(CalibrationMaker):

    def __init__(self, pipeline_context):
        super(BiasMaker, self).__init__(pipeline_context)

    @property
    def group_by_keywords(self):
        return ['ccdsum']

    @property
    def calibration_type(self):
        return 'BIAS'

    @property
    def min_images(self):
        return 5

    def make_master_calibration_frame(self, images, image_config, logging_tags):

        bias_data = np.zeros((image_config.ny, image_config.nx, len(images)), dtype=np.float32)
        bias_mask = np.zeros((image_config.ny, image_config.nx, len(images)), dtype=np.uint8)
        bias_level_array = np.zeros(len(images), dtype=np.float32)

        master_bias_filename = self.get_calibration_filename(image_config)
        logs.add_tag(logging_tags, 'master_bias', os.path.basename(master_bias_filename))
        for i, image in enumerate(images):
            bias_level_array[i] = stats.sigma_clipped_mean(image.data, 3.5, mask=image.bpm)

            logs.add_tag(logging_tags, 'filename', os.path.basename(image.filename))
            logs.add_tag(logging_tags, 'BIASLVL', float(bias_level_array[i]))
            self.logger.debug('Calculating bias level', extra=logging_tags)
            # Subtract the bias level for each image
            bias_data[:, :, i] = image.data[:, :] - bias_level_array[i]
            bias_mask[:, :, i] = image.bpm[:, :]

        mean_bias_level = stats.sigma_clipped_mean(bias_level_array, 3.0)

        master_bias = stats.sigma_clipped_mean(bias_data, 3.0, axis=2, mask=bias_mask, inplace=True)

        del bias_data
        del bias_mask

        master_bpm = np.array(master_bias == 0.0, dtype=np.uint8)

        header = fits_utils.create_master_calibration_header(images)

        header['BIASLVL'] = (mean_bias_level, 'Mean bias level of master bias')
        master_bias_image = Image(data=master_bias, header=header)
        master_bias_image.filename = master_bias_filename
        master_bias_image.bpm = master_bpm

        logs.pop_tag(logging_tags, 'master_bias')
        logs.add_tag(logging_tags, 'filename', os.path.basename(master_bias_image.filename))
        logs.add_tag(logging_tags, 'BIASLVL', mean_bias_level)
        self.logger.debug('Average bias level in ADU', extra=logging_tags)

        return [master_bias_image]


class BiasSubtractor(ApplyCalibration):
    def __init__(self, pipeline_context):
        super(BiasSubtractor, self).__init__(pipeline_context)

    @property
    def group_by_keywords(self):
        return ['ccdsum']

    @property
    def calibration_type(self):
        return 'bias'

    def apply_master_calibration(self, images, master_calibration_image, logging_tags):

        master_bias_data = master_calibration_image.data
        master_bias_level = float(master_calibration_image.header['BIASLVL'])

        logs.add_tag(logging_tags, 'bias', master_bias_level)
        logs.add_tag(logging_tags, 'master_bias',
                     os.path.basename(master_calibration_image.filename))

        for image in images:
            logs.add_tag(logging_tags, 'filename', os.path.basename(image.filename))

            image.subtract(master_bias_level)
            image.subtract(master_bias_data)

            image.bpm |= master_calibration_image.bpm
            image.header['BIASLVL'] = (master_bias_level, 'Mean bias level of master bias')

            master_bias_filename = os.path.basename(master_calibration_image.filename)
            image.header['L1IDBIAS'] = (master_bias_filename, 'ID of bias frame used')
            image.header['L1STATBI'] = (1, "Status flag for bias frame correction")
            logs.add_tag(logging_tags, 'BIASLVL', image.header['BIASLVL'])
            logs.add_tag(logging_tags, 'L1IDBIAS', image.header['L1IDBIAS'])

            self.logger.info('Subtracting bias', extra=logging_tags)
        return images


class OverscanSubtractor(Stage):
    def __init__(self, pipeline_context):
        super(OverscanSubtractor, self).__init__(pipeline_context)

    @property
    def group_by_keywords(self):
        return None

    def do_stage(self, images):

        for image in images:
            logging_tags = logs.image_config_to_tags(image, self.group_by_keywords)
            logs.add_tag(logging_tags, 'filename', os.path.basename(image.filename))

            # Subtract the overscan if it exists
            if len(image.data.shape) > 2:
                for i in range(image.data.shape[0]):
                    overscan_level = _subtract_overscan_3d(image, i)
                    logs.add_tag(logging_tags, 'OVERSCN{0}'.format(i + 1), overscan_level)
                    self.logger.info('Subtracting overscan', extra=logging_tags)
            else:
                overscan_level = _subtract_overscan_2d(image)
                logs.add_tag(logging_tags, 'OVERSCAN', overscan_level)
                self.logger.info('Subtracting overscan', extra=logging_tags)

        return images


def _subtract_overscan_3d(image, i):
    overscan_region = fits_utils.parse_region_keyword(image.header.get('BIASSEC{0}'.format(i + 1)))
    if overscan_region is not None:
        overscan_level = stats.sigma_clipped_mean(image.data[i][overscan_region], 3)
        image.header['L1STATOV'] = (1, 'Status flag for overscan correction')
    else:
        overscan_level = 0.0
        image.header['L1STATOV'] = (0, 'Status flag for overscan correction')

    overscan_comment = 'Overscan value that was subtracted from Q{0}'.format(i + 1)
    image.header['OVERSCN{0}'.format(i + 1)] = (overscan_level, overscan_comment)

    image.data[i] -= overscan_level
    return overscan_level


def _subtract_overscan_2d(image):
    overscan_region = fits_utils.parse_region_keyword(image.header.get('BIASSEC'))
    if overscan_region is not None:
        overscan_level = stats.sigma_clipped_mean(image.data[overscan_region], 3)
        image.header['L1STATOV'] = (1, 'Status flag for overscan correction')
    else:
        overscan_level = 0.0
        image.header['L1STATOV'] = (0, 'Status flag for overscan correction')

    image.header['OVERSCAN'] = (overscan_level, 'Overscan value that was subtracted')
    image.data -= overscan_level
    return overscan_level



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
import os
import sys
import subprocess

from distutils.core import Extension
from distutils import log

from astropy_helpers import setup_helpers

UTIL_DIR = os.path.relpath(os.path.dirname(__file__))

CODELINES = r"""
import sys
import os
from distutils.ccompiler import new_compiler
ccompiler = new_compiler()
ccompiler.add_library('gomp')
has_omp_functions = ccompiler.has_function('omp_get_num_threads')
with open('openmp_check.c', 'w') as f:
    f.write('#include<stdio.h>\n')
    f.write('main()\n')
    f.write('{\n')
    f.write('printf("Hello World");')
    f.write('}')
try:
    import pdb; pdb.set_trace()
    ccompiler.compile(['openmp_check.c'], extra_postargs=['-fopenmp'])
    fopenmp_flag_works = True
except:
    fopenmp_flag_works = False
os.remove('openmp_check.c')
if os.path.exists('openmp_check.o'):
    os.remove('openmp_check.o')
sys.exit(int(has_omp_functions & fopenmp_flag_works))
"""


def check_openmp():
    if setup_helpers.get_compiler_option() == 'msvc':
        # The free version of the Microsoft compilers supports
        # OpenMP in MSVC 2008 (python 2.7) and MSVC 2015 (python 3.5+),
        # but not MSVC 2010 (python 3.4 and lower).
        major, minor = sys.version_info[:2]
        has_openmp = not (major == 3 and minor < 5)
        # Empty return tuple is to match the alternative check, below.
        return has_openmp, ("", "")
    else:
        # Unix-y compiler, use this check.
        s = subprocess.Popen([sys.executable], stdin=subprocess.PIPE,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE)
        stdout, stderr = s.communicate(CODELINES.encode('utf-8'))
        s.wait()
        return bool(s.returncode), (stdout, stderr)


def get_extensions():

    med_sources = [str(os.path.join(UTIL_DIR, "median_utils.pyx")),
                   str(os.path.join(UTIL_DIR, "quick_select.c"))]

    include_dirs = ['numpy', UTIL_DIR]

    libraries = []

    ext_med = Extension(name=str('banzai.utils.median_utils'),
                        sources=med_sources,
                        include_dirs=include_dirs,
                        libraries=libraries,
                        language="c",
                        extra_compile_args=['-g', '-O3', '-funroll-loops', '-ffast-math'])

    has_openmp, outputs = check_openmp()
    if has_openmp:
        if setup_helpers.get_compiler_option() == 'msvc':
            ext_med.extra_compile_args.append('-openmp')
        else:
            ext_med.extra_compile_args.append('-fopenmp')
            ext_med.extra_link_args = ['-g', '-fopenmp']
    else:
        log.warn('OpenMP was not found. '
                 'banzai will be compiled without OpenMP. '
                 '(Use the "-v" option of setup.py for more details.)')
        log.debug(('(Start of OpenMP info)\n'
                   'compiler stdout:\n{0}\n'
                   'compiler stderr:\n{1}\n'
                   '(End of OpenMP info)').format(*outputs))

    return [ext_med]



""" dbs.py: Database utility functions for banzai

    This is built around the SQLAlchemy ORM

Author
    Curtis McCully (cmccully@lcogt.net)

October 2015
"""
from __future__ import absolute_import, print_function, division

import os.path

from sqlalchemy import create_engine, pool
from sqlalchemy.orm import sessionmaker
from sqlalchemy import Column, Integer, String, Float, Date, ForeignKey, DateTime, Boolean, CHAR
from sqlalchemy.ext.declarative import declarative_base

from glob import glob
from astropy.io import fits
import requests
from banzai.utils import file_utils


# Define how to get to the database
# Note that we need to encode the database password outside of the code base
_DEFAULT_DB = 'mysql://cmccully:password@localhost/test'

Base = declarative_base()


def get_session(db_address=_DEFAULT_DB):
    """
    Get a connection to the database.

    Returns
    -------
    session: SQLAlchemy Database Session
    """
    # Build a new engine for each session. This makes things thread safe.
    engine = create_engine(db_address, poolclass=pool.NullPool)
    Base.metadata.bind = engine

    # We don't use autoflush typically. I have run into issues where SQLAlchemy would try to flush
    # incomplete records causing a crash. None of the queries here are large, so it should be ok.
    db_session = sessionmaker(bind=engine, autoflush=False)
    session = db_session()

    return session


class Image(Base):
    """
    Image Database Record

    This defines the images table. Most of these keywords are parsed from the headers.
    telescope_id is a foreign key to the telescopes table.
    """
    __tablename__ = 'images'

    # Define the table structure
    id = Column(Integer, primary_key=True, autoincrement=True)
    filename = Column(String(36), index=True, unique=True)
    telescope_id = Column(Integer, ForeignKey("telescopes.id"), index=True)
    filepath = Column(String(100))
    rawfilename = Column(String(36))
    rawpath = Column(String(100))
    object_name = Column(String(50))
    mjd = Column(Float, index=True)
    dateobs = Column(DateTime)
    dayobs = Column(Date, index=True)
    exptime = Column(Float)
    filter_name = Column(String(6))
    obstype = Column(String(20))
    airmass = Column(Float)
    ra = Column(Float)
    dec = Column(Float)
    userid = Column(String(20))
    propid = Column(String(20))
    tracknum = Column(String(20))
    reqnum = Column(String(20))
    ccdsum = Column(String(10))
    gain = Column(Float)
    readnoise = Column(Float)
    naxis1 = Column(Integer)
    naxis2 = Column(Integer)
    pixel_scale = Column(Float)
    focus = Column(Integer)
    # Reduction Status
    ingest_done = Column(Boolean, default=False)
    bias_done = Column(Boolean, default=False)
    trim_done = Column(Boolean, default=False)
    dark_done = Column(Boolean, default=False)
    flat_done = Column(Boolean, default=False)
    wcs_done = Column(Boolean, default=False)
    cat_done = Column(Boolean, default=False)

    def get_full_filename(self, suffix, extension='.fits'):
        image_file = os.path.join(self.filepath, self.filename)
        image_file += suffix + extension

        return image_file


class CalibrationImage(Base):
    """
    Master Calibration Image Database Record

    This defines the calimages table. We use this to keep track of the master calibration frames.
    Typically these are bias, darks, and flat field frames. These are indexed by dayobs to make it
    easy to find the closest calibration frame.
    """
    __tablename__ = 'calimages'
    id = Column(Integer, primary_key=True, autoincrement=True)
    type = Column(String(30), index=True)
    filename = Column(String(50), unique=True)
    filepath = Column(String(100))
    dayobs = Column(Date, index=True)
    ccdsum = Column(String(20))
    filter_name = Column(String(6))
    telescope_id = Column(Integer, ForeignKey("telescopes.id"), index=True)


class Telescope(Base):
    """
    Telescope Database Record

    This defines the telescopes table.
    """
    __tablename__ = 'telescopes'
    id = Column(Integer, primary_key=True, autoincrement=True)
    site = Column(String(10), index=True)
    instrument = Column(String(20), index=True)
    camera_type = Column(String(20))


class BadPixelMask(Base):
    """
    Bad Pixel Mask Database Record
    """
    __tablename__ = 'bpms'
    id = Column(Integer, primary_key=True, autoincrement=True)
    telescope_id = Column(Integer, ForeignKey("telescopes.id"), index=True)
    filename = Column(String(50))
    filepath = Column(String(100))
    ccdsum = Column(String(20))


class PreviewImage(Base):
    __tablename__ = 'previewimages'
    id = Column(Integer, primary_key=True, autoincrement=True)
    filename = Column(String(50), index=True)
    checksum = Column(CHAR(32), index=True)


def create_db(bpm_directory, db_address=_DEFAULT_DB,
              configdb_address='http://configdb.lco.gtn/sites/'):
    """
    Create the database structure.

    This only needs to be run once on initialization of the database.
    """
    # Create an engine for the database
    engine = create_engine(db_address)

    # Create all tables in the engine
    # This only needs to be run once on initialization.
    Base.metadata.create_all(engine)

    populate_telescope_table(db_address)
    populate_bpm_table(bpm_directory, db_address=db_address, configdb_address=configdb_address)


def parse_configdb(configdb_address='http://configdb.lco.gtn/sites/'):
    """
    Parse the contents of the configdb.

    Parameters
    ----------
    configdb_address : str
                      URL of the configdb, must be inside LCOGT VPN

    Returns
    -------
    cameras : list of dicts
              each camera dictionary contains a site, instrument code, and camera type.
    """
    sites = requests.get(configdb_address).json()['results']
    cameras = []
    for site in sites:
        for enc in site['enclosure_set']:
            for tel in enc['telescope_set']:
                for ins in tel['instrument_set']:
                    sci_cam = ins.get('science_camera')
                    if sci_cam is not None:
                        if 'SciCam' in sci_cam['camera_type']['code']:
                            cameras.append({'site': site['code'],
                                            'instrument': sci_cam['code'],
                                            'camera_type': sci_cam['camera_type']['code']})
    return cameras


def populate_telescope_table(db_address=_DEFAULT_DB,
                             configdb_address='http://configdb.lco.gtn/sites/'):
    """
    Populate the telescope table

    Parameters
    ----------
    db_address : str
                 sqlalchemy address to the database of the form
                 mysql://username:password@localhost/test

    configdb_address : str
                       URL of the configdb

    Notes
    -----
    This only works inside the LCOGT VPN. This should be run at least when a new camera is
    added to the network
    """

    cameras = parse_configdb(configdb_address=configdb_address)

    db_session = get_session(db_address=db_address)

    for camera in cameras:

        # Check and see if the site and instrument combinatation already exists in the table
        camera_query = Telescope.site == camera['site']
        camera_query &= Telescope.instrument == camera['instrument']
        matching_cameras = db_session.query(Telescope).filter(camera_query).all()

        if len(matching_cameras) == 0:
            db_session.add(Telescope(site=camera['site'], instrument=camera['instrument'],
                                     camera_type=camera['camera_type']))

    db_session.commit()
    db_session.close()


def populate_bpm_table(directory, db_address=_DEFAULT_DB):
    db_session = get_session(db_address=db_address)
    bpm_filenames = glob(os.path.join(directory, 'bpm*.fits'))
    for bpm_filename in bpm_filenames:
        site = fits.getval(bpm_filename, 'SITEID').lower()
        instrument = fits.getval(bpm_filename, 'INSTRUME').lower()
        ccdsum = fits.getval(bpm_filename, 'CCDSUM')

        telescope_query = Telescope.site == site
        telescope_query &= Telescope.instrument == instrument
        telescope = db_session.query(Telescope).filter(telescope_query).first()

        if telescope is not None:

            db_session.add(BadPixelMask(telescope_id=telescope.id, filepath=os.path.abspath(directory),
                                        filename=os.path.basename(bpm_filename), ccdsum=ccdsum))

    db_session.commit()
    db_session.close()


def get_telescope_id(site, instrument, db_address=_DEFAULT_DB):
    # TODO:  This dies if the telescope is not in the telescopes table. Maybe ping the configdb?
    db_session = get_session(db_address=db_address)
    criteria = (Telescope.site == site) & (Telescope.instrument == instrument)
    telescope = db_session.query(Telescope).filter(criteria).first()
    db_session.close()
    return telescope.id


def get_telescope(telescope_id, db_address=_DEFAULT_DB):
    db_session = get_session(db_address=db_address)
    telescope = db_session.query(Telescope).filter(Telescope.id == telescope_id).first()
    db_session.close()
    return telescope


def get_bpm(telescope_id, ccdsum, db_address=_DEFAULT_DB):
    db_session = get_session(db_address=db_address)
    bpm = db_session.query(BadPixelMask).filter(BadPixelMask.telescope_id == telescope_id,
                                                BadPixelMask.ccdsum == ccdsum).first()
    db_session.close()
    if bpm is not None:
        bpm_path = os.path.join(bpm.filepath, bpm.filename)
    else:
        bpm_path = None
    return bpm_path


def save_calibration_info(cal_type, output_file, image_config, db_address=_DEFAULT_DB):
    # Store the information into the calibration table
    # Check and see if the bias file is already in the database
    db_session = get_session(db_address=db_address)
    image_query = db_session.query(CalibrationImage)
    output_filename = os.path.basename(output_file)
    image_query = image_query.filter(CalibrationImage.filename == output_filename)
    image_query = image_query.all()

    if len(image_query) == 0:
        # Create a new row
        calibration_image = CalibrationImage()
    else:
        # Otherwise update the existing data
        # In principle we could just skip this, but this should be fast
        calibration_image = image_query[0]

    calibration_image.dayobs = image_config.epoch
    calibration_image.ccdsum = image_config.ccdsum
    calibration_image.filter_name = image_config.filter
    calibration_image.telescope_id = image_config.telescope_id
    calibration_image.type = cal_type.upper()
    calibration_image.filename = output_filename
    calibration_image.filepath = os.path.dirname(output_file)

    db_session.add(calibration_image)
    db_session.commit()
    db_session.close()


def preview_file_already_processed(path, db_address=_DEFAULT_DB):
    db_session = get_session(db_address=db_address)
    query = db_session.query(PreviewImage)
    criteria = PreviewImage.filename == os.path.basename(path)
    criteria &= PreviewImage.checksum == file_utils.get_md5(path)
    query = query.filter(criteria)
    already_processed = len(query.all()) > 0
    db_session.close()
    return already_processed


def set_preview_file_as_processed(path, db_address=_DEFAULT_DB):
    filename = os.path.basename(path)
    checksum = file_utils.get_md5(path)

    db_session = get_session(db_address=db_address)
    query = db_session.query(PreviewImage)
    criteria = PreviewImage.filename == filename
    criteria &= PreviewImage.checksum == checksum
    query = query.filter(criteria)
    if len(query.all()) == 0:
        preview_image = PreviewImage(filename=filename, checksum=checksum)
        db_session.add(preview_image)
        db_session.commit()
    db_session.close()
